# -*- coding: utf-8 -*-

"""
Tests for Aho-Corasick string search algorithm.
Original Author: Wojciech Muła, wojciech_mula@poczta.onet.pl
WWW            : http://0x80.pl
License        : public domain

Modified for use in the license_expression library and in particular:
 - add support for unicode key strinsg.
 - rename word to key and output to value (to be more like a mapping/dict)
 - case insensitive search
 - improve returned results with the actual start,end and matched string.
 - support returning non-matched parts of a string
"""

from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import print_function

import unittest

from license_expression._pyahocorasick import Trie
from license_expression._pyahocorasick import Output
from license_expression._pyahocorasick import Result


class TestTrie(unittest.TestCase):
    def testAddedWordShouldBeCountedAndAvailableForRetrieval(self):
        t = Trie()
        t.add('python', 'value')
        assert Output('python', 'value') == t.get('python')

    def testAddingExistingWordShouldReplaceAssociatedValue(self):
        t = Trie()
        t.add('python', 'value')
        assert Output('python', 'value') == t.get('python')

        t.add('python', 'other')
        assert Output('python', 'other') == t.get('python')

    def testGetUnknowWordWithoutDefaultValueShouldRaiseException(self):
        t = Trie()
        with self.assertRaises(KeyError):
            t.get('python')

    def testGetUnknowWordWithDefaultValueShouldReturnDefault(self):
        t = Trie()
        self.assertEqual(t.get('python', 'default'), 'default')

    def testExistShouldDetectAddedWords(self):
        t = Trie()
        t.add('python', 'value')
        t.add('ada', 'value')

        self.assertTrue(t.exists('python'))
        self.assertTrue(t.exists('ada'))

    def testExistShouldReturnFailOnUnknownWord(self):
        t = Trie()
        t.add('python', 'value')

        self.assertFalse(t.exists('ada'))

    def test_is_prefix_ShouldDetecAllPrefixesIncludingWord(self):
        t = Trie()
        t.add('python', 'value')
        t.add('ada', 'value')

        self.assertTrue(t.is_prefix('a'))
        self.assertTrue(t.is_prefix('ad'))
        self.assertTrue(t.is_prefix('ada'))

        self.assertTrue(t.is_prefix('p'))
        self.assertTrue(t.is_prefix('py'))
        self.assertTrue(t.is_prefix('pyt'))
        self.assertTrue(t.is_prefix('pyth'))
        self.assertTrue(t.is_prefix('pytho'))
        self.assertTrue(t.is_prefix('python'))

    def testItemsShouldReturnAllItemsAlreadyAddedToTheTrie(self):
        t = Trie()

        t.add('python', 1)
        t.add('ada', 2)
        t.add('perl', 3)
        t.add('pascal', 4)
        t.add('php', 5)

        result = list(t.items())
        self.assertIn(('python', 1), result)
        self.assertIn(('ada', 2), result)
        self.assertIn(('perl', 3), result)
        self.assertIn(('pascal', 4), result)
        self.assertIn(('php', 5), result)


    def testKeysShouldReturnAllKeysAlreadyAddedToTheTrie(self):
        t = Trie()

        t.add('python', 1)
        t.add('ada', 2)
        t.add('perl', 3)
        t.add('pascal', 4)
        t.add('php', 5)

        result = list(t.keys())
        self.assertIn('python', result)
        self.assertIn('ada', result)
        self.assertIn('perl', result)
        self.assertIn('pascal', result)
        self.assertIn('php', result)


    def testValuesShouldReturnAllValuesAlreadyAddedToTheTrie(self):
        t = Trie()

        t.add('python', 1)
        t.add('ada', 2)
        t.add('perl', 3)
        t.add('pascal', 4)
        t.add('php', 5)

        result = list(t.values())
        self.assertIn(1, result)
        self.assertIn(2, result)
        self.assertIn(3, result)
        self.assertIn(4, result)
        self.assertIn(5, result)

    def test_iter_should_not_return_non_matches(self):

        def get_test_automaton():
            words = "he her hers his she hi him man himan".split()
            t = Trie()
            for w in words:
                t.add(w, w)
            t.make_automaton()
            return t

        test_string = "he she himan"

        t = get_test_automaton()
        result = list(t.iter(test_string))
        expected = [
            Result(start=0, end=1, string='he', output=Output('he', 'he')),
            Result(start=3, end=5, string='she', output=Output('she', 'she')),
            Result(start=4, end=5, string='he', output=Output('he', 'he')),
            Result(start=7, end=8, string='hi', output=Output('hi', 'hi')),
            Result(start=7, end=9, string='him', output=Output('him', 'him')),
            Result(start=7, end=11, string='himan', output=Output('himan', 'himan')),
            Result(start=9, end=11, string='man', output=Output('man', 'man'))
        ]

        assert expected == result

    def test_iter_vs_scan(self):
        def get_test_automaton():
            words = "( AND ) OR".split()
            t = Trie()
            for w in words:
                t.add(w, w)
            t.make_automaton()
            return t

        test_string = '((l-a + AND l-b) OR (l -c+))'

        t = get_test_automaton()
        result = list(t.iter(test_string))
        expected = [
            Result(0, 0, '(', Output('(', '(')),
            Result(1, 1, '(', Output('(', '(')),
            Result(8, 10, 'AND', Output('AND', 'AND')),
            Result(15, 15, ')', Output(')', ')')),
            Result(17, 18, 'OR', Output('OR', 'OR')),
            Result(20, 20, '(', Output('(', '(')),
            Result(26, 26, ')', Output(')', ')')),
            Result(27, 27, ')', Output(')', ')'))
        ]
        assert expected == result

        result = list(t.scan(test_string))
        expected = [
            Result(0, 0, '(', Output('(', '(')),
            Result(1, 1, '(', Output('(', '(')),
            Result(2, 7, 'l-a + ', None),
            Result(8, 10, 'AND', Output('AND', 'AND')),
            Result(11, 14, ' l-b', None),
            Result(15, 15, ')', Output(')', ')')),
            Result(16, 16, ' ', None),
            Result(17, 18, 'OR', Output('OR', 'OR')),
            Result(19, 19, ' ', None),
            Result(20, 20, '(', Output('(', '(')),
            Result(21, 25, 'l -c+', None),
            Result(26, 26, ')', Output(')', ')')),
            Result(27, 27, ')', Output(')', ')'))
        ]
        assert expected == result

    def test_scan_with_unmatched(self):
        def get_test_automaton():
            words = "( AND ) OR".split()
            t = Trie()
            for w in words:
                t.add(w, w)
            t.make_automaton()
            return t

        test_string = '((l-a + AND l-b) OR an (l -c+))'

        t = get_test_automaton()
        result = list(t.scan(test_string))
        assert test_string == ''.join(r.string for r in result)

    def test_iter_with_unmatched_simple(self):
        t = Trie()
        t.add('AND', 'AND')
        t.make_automaton()
        test_string = 'AND  an a and'
        result = list(t.iter(test_string))
        assert 'ANDand' == ''.join(r.string for r in result)

    def test_iter_with_unmatched_simple2(self):
        t = Trie()
        t.add('AND', 'AND')
        t.make_automaton()
        test_string = 'AND  an a and'
        result = list(t.iter(test_string))
        assert 'ANDand' == ''.join(r.string for r in result)




# -*- coding: utf-8 -*-
"""
Aho-Corasick string search algorithm.

Original Author: Wojciech Muła, wojciech_mula@poczta.onet.pl
WWW            : http://0x80.pl
License        : public domain

Modified for use in the license_expression library and in particular:
 - add support for unicode key strinsg.
 - rename word to key and output to value (to be more like a mapping/dict)
 - case insensitive search
 - improve returned results with the actual start,end and matched string.
 - support returning non-matched parts of a string
"""

from __future__ import unicode_literals
from __future__ import absolute_import
from __future__ import print_function

from collections import deque
from collections import OrderedDict
import logging

logger = logging.getLogger(__name__)

def logger_debug(*args):
    return logger.debug(' '.join(isinstance(a, str) and a or repr(a) for a in args))

# uncomment for local debug logging
# import sys
# logging.basicConfig(stream=sys.stdout)
# logger.setLevel(logging.DEBUG)


# used to distinguish from None
nil = object()


class Trie(object):
    """
    A Trie and Aho-Corasick automaton. This behaves more or less like a mapping of
    key->value. This is the main entry point.
    """
    def __init__(self, ignore_case=True):
        """
        Initialize a new Trie.

        If `ignore_case` is True, searches in the Trie will be case insensitive.
        """
        self.root = TrieNode('')
        self.ignore_case = ignore_case

        # set of any unique character in the trie, updated on each addition
        # we keep track of the set of chars added to the trie to build the automaton
        # these are needed to created the first level children failure links
        self._known_chars = set()

        # Flag set to True once a Trie has been converted to an Aho-Corasick automaton
        self._converted = False

    def add(self, key, value=None):
        """
        Add a new (key string, value) pair to the trie. If the key already exists in
        the Trie, its value is replaced with the provided value.
        A key is any unicode string.
        """
        if self._converted:
            raise Exception('This Trie has been converted to an Aho-Corasick '
                            'automaton and cannot be further modified.')
        if not key:
            return

        stored_key = self.ignore_case and key.lower() or key

        # we keep track of the set of chars added to the trie to build the automaton
        # these are needed to created the first level children failure links
        self._known_chars.update(stored_key)

        node = self.root
        for char in stored_key:
            try:
                node = node.children[char]
            except KeyError:
                child = TrieNode(char)
                node.children[char] = child
                node = child

        # we always store the original key, not a possibly lowercased version
        node.output = Output(key, value)

    def __get_node(self, key):
        """
        Return a node for this key or None if the trie does not contain the key.
        Private function retrieving a final node of trie for given key.
        """
        key = self.ignore_case and key.lower() or key
        node = self.root
        for char in key:
            try:
                node = node.children[char]
            except KeyError:
                return None
        return node

    def get(self, key, default=nil):
        """
        Return the Output tuple associated with a `key`.
        If there is no such key in the Trie, return the default value (other
        than nil): if default is not given or nil, raise a KeyError exception.
        """
        node = self.__get_node(key)
        output = nil
        if node:
            output = node.output

        if output is nil:
            if default is nil:
                raise KeyError(key)
            else:
                return default
        else:
            return output

    def keys(self):
        """
        Yield all keys stored in this trie.
        """
        return (key for key, _ in self.items())

    def values(self):
        """
        Yield all values associated with keys stored in this trie.
        """
        return (value for _, value in self.items())

    def items(self):
        """
        Yield tuple of all (key, value) stored in this trie.
        """
        items = []

        def walk(node, key):
            """
            Walk the trie, depth first.
            """
            key = key + node.char
            if node.output is not nil:
                items.append((node.output.key, node.output.value))

            for child in node.children.values():
                if child is not node:
                    walk(child, key)

        walk(self.root, key='')

        return iter(items)

    def exists(self, key):
        """
        Return True if the key is present in this trie.
        """
        # TODO: add __contains__ magic for this
        node = self.__get_node(key)
        if node:
            return bool(node.output != nil)
        return False

    def is_prefix(self, key):
        """
        Return True if key is a prefix of any existing key in the trie.
        """
        return (self.__get_node(key) is not None)

    def make_automaton(self):
        """
        Convert this trie to an Aho-Corasick automaton.
        Note that this is an error to add new keys to a Trie once it has been
        converted to an Automaton.
        """
        queue = deque()
        queue_append = queue.append
        queue_popleft = queue.popleft

        # 1. create root children for each known items range (e.g. all unique
        # characters from all the added keys), failing to root.
        # And build a queue of these
        for char in self._known_chars:
            if char in self.root.children:
                node = self.root.children[char]
                # e.g. f(s) = 0, Aho-Corasick-wise
                node.fail = self.root
                queue_append(node)
            else:
                self.root.children[char] = self.root

        # 2. using the queue of all possible top level items/chars, walk the trie and
        # add failure links to nodes as needed
        while queue:
            current_node = queue_popleft()
            for node in current_node.children.values():
                queue_append(node)
                state = current_node.fail
                while node.char not in state.children:
                    state = state.fail
                node.fail = state.children.get(node.char, self.root)

        # Mark the trie as converted so it cannot be modified anymore
        self._converted = True

    def iter(self, string):
        """
        Yield Result objects for matched strings by performing the Aho-Corasick search procedure.

        The Result start and end positions in the searched string are such that the
        matched string is "search_string[start:end+1]". And the start is computed
        from the end_index collected by the Aho-Corasick search procedure such that
        "start=end_index - n + 1" where n is the length of a matched key.

        The Result.output is an Output object for a matched key.

        For example:
        >>> a = Trie()
        >>> a.add('BCDEF')
        >>> a.add('CDE')
        >>> a.add('DEFGH')
        >>> a.add('EFGH')
        >>> a.add('KL')
        >>> a.make_automaton()
        >>> string = 'abcdefghijklm'
        >>> results = Result.sort(a.iter(string))

        >>> expected = [
        ...     Result(1, 5, 'bcdef', Output('BCDEF')),
        ...     Result(2, 4, 'cde', Output('CDE')),
        ...     Result(3, 7, 'defgh', Output('DEFGH')),
        ...     Result(4, 7, 'efgh', Output('EFGH')),
        ...     Result(10, 11, 'kl', Output('KL')),
        ... ]
        >>> results == expected
        True

        >>> list(a.iter('')) == []
        True

        >>> list(a.iter(' ')) == []
        True
        """
        if not string:
            return

        # keep a copy for results
        original_string = string
        string = self.ignore_case and string.lower() or string

        known_chars = self._known_chars
        state = self.root
        for end, char in enumerate(string):
            if char not in known_chars:
                state = self.root
                continue

            # search for a matching character in the children, starting at root
            while char not in state.children:
                state = state.fail
            # we have a matching starting character
            state = state.children.get(char, self.root)
            match = state
            while match is not nil:
                if match.output is not nil:
                    # TODO: this could be precomputed or cached
                    n = len(match.output.key)
                    start = end - n + 1
                    yield Result(start, end, original_string[start:end + 1], match.output)
                match = match.fail

    def scan(self, string):
        """
        Scan a string for matched and unmatched sub-sequences and yield non-
        overlapping Result objects performing a modified Aho-Corasick search
        procedure:

        - return both matched and unmatched sub-sequences.
        - do not return matches with positions that are contained or overlap with
          another match:
          - discard smaller matches contained in a larger match.
          - when there is overlap (but not containment), the matches are sorted by
            start and biggest length and then:
             - we return the largest match of two overlaping matches
             - if they have the same length, keep the match starting the earliest and
               return the non-overlapping portion of the other discarded match as a
               non-match.

        Each Result contains the start and end position, the corresponding string and
        an Output object (with original key and any associated associated value). The
        string and key are in their original case even if the automaton has the
        `ignore_case` attribute.

        For example:
        >>> a = Trie()
        >>> a.add('BCDEF')
        >>> a.add('CDE')
        >>> a.add('DEFGH')
        >>> a.add('EFGH')
        >>> a.add('KL')
        >>> a.make_automaton()
        >>> string = 'abcdefghijkl'
        >>> results = list(a.scan(string))

        >>> expected = [
        ...     Result(start=0, end=0, string='a', output=None),
        ...     Result(start=1, end=5, string='bcdef', output=Output('BCDEF')),
        ...     Result(start=6, end=9, string='ghij', output=None),
        ...     Result(start=10, end=11, string='kl', output=Output('KL')),
        ... ]

        >>> results == expected
        True
        """
        results = self.iter(string)
        results = filter_overlapping(results)
        results = add_unmatched(string, results)
        return results


class TrieNode(object):
    """
    Node of the Trie/Aho-Corasick automaton.
    """
    __slots__ = ['char', 'output', 'fail', 'children']

    def __init__(self, char, output=nil):
        # character of a key string added to the Trie
        self.char = char

        # an output function (in the Aho-Corasick meaning) for this node: this is an
        # Output object that contains the original key string and any additional
        # value data associated to that key. Or "nil" for a node that is not a
        # terminal leave for a key. It will be returned with a match.
        self.output = output

        # failure link used by the Aho-Corasick automaton and its search procedure
        self.fail = nil

        # children of this node as a mapping of char->node
        self.children = {}

    def __repr__(self):
        if self.output is not nil:
            return 'TrieNode(%r, %r)' % (self.char, self.output)
        else:
            return 'TrieNode(%r)' % self.char


class Output(object):
    """
    An Output is used to track a key added to the Trie as a TrieNode and any
    arbitrary value object corresponding to that key.

    - `key` is the original key unmodified unicode string.
    - `value` is the associated value for this key as provided when adding this key.
    """
    __slots__ = 'key', 'value'

    def __init__(self, key, value=None):
        self.key = key
        self.value = value

    def __repr__(self):
        return self.__class__.__name__ + '(%(key)r, %(value)r)' % self.as_dict()

    def __eq__(self, other):
        return isinstance(other, Output) and self.key == other.key and self.value == other.value

    def __hash__(self):
        return hash((self.key, self.value,))

    def as_dict(self):
        return OrderedDict([(s, getattr(self, s)) for s in self.__slots__])


class Result(object):
    """
    A Result is used to track the result of a search with its start and end as
    index position in the original string and other attributes:

    - `start` and `end` are zero-based index in the original string S such that
         S[start:end+1] will yield `string`.
    - `string` is the sub-string from the original searched string for this Result.
    - `output` is the Output object for a matched string and is a marker that this is a
       matched string. None otherwise for a Result for unmatched text.
    """

    __slots__ = 'start', 'end', 'string', 'output'

    def __init__(self, start, end, string='', output=None):
        self.start = start
        self.end = end
        self.string = string
        self.output = output

    def __repr__(self):
        return self.__class__.__name__ + '(%(start)r, %(end)r, %(string)r, %(output)r)' % self.as_dict()

    def as_dict(self):
        return OrderedDict([(s, getattr(self, s)) for s in self.__slots__])

    def __len__(self):
        return self.end + 1 - self.start

    def __eq__(self, other):
        return isinstance(other, Result) and (
            self.start == other.start and
            self.end == other.end and
            self.string == other.string and
            self.output == other.output
        )

    def __hash__(self):
        tup = self.start, self.end, self.string, self.output
        return hash(tup)


    def is_after(self, other):
        """
        Return True if this result is after the other result.

        For example:
        >>> Result(1, 2).is_after(Result(5, 6))
        False
        >>> Result(5, 6).is_after(Result(5, 6))
        False
        >>> Result(2, 3).is_after(Result(1, 2))
        False
        >>> Result(5, 6).is_after(Result(3, 4))
        True
        """
        return self.start > other.end

    def is_before(self, other):
        return self.end < other.start

    def __contains__(self, other):
        """
        Return True if this result contains the other result.

        For example:
        >>> Result(5, 7) in Result(5, 7)
        True
        >>> Result(6, 8) in Result(5, 7)
        False
        >>> Result(6, 6) in Result(4, 8)
        True
        >>> Result(3, 9) in Result(4, 8)
        False
        >>> Result(4, 8) in Result(3, 9)
        True
        """
        return self.start <= other.start and other.end <= self.end

    def overlap(self, other):
        """
        Return True if this result and the other result overlap.

        For example:
        >>> Result(1, 2).overlap(Result(5, 6))
        False
        >>> Result(5, 6).overlap(Result(5, 6))
        True
        >>> Result(4, 5).overlap(Result(5, 6))
        True
        >>> Result(4, 5).overlap(Result(5, 7))
        True
        >>> Result(4, 5).overlap(Result(6, 7))
        False
        """
        start = self.start
        end = self.end
        return (start <= other.start <= end) or (start <= other.end <= end)

    @classmethod
    def sort(cls, results):
        """
        Return a new sorted sequence of results given a sequence of results. The
        primary sort is on start and the secondary sort is on longer lengths.
        Therefore if two results have the same start, the longer result will sort
        first.

        For example:
        >>> results = [Result(0, 0), Result(5, 5), Result(1, 1), Result(2, 4), Result(2, 5)]
        >>> expected = [Result(0, 0), Result(1, 1), Result(2, 5), Result(2, 4), Result(5, 5)]
        >>> expected == Result.sort(results)
        True
        """
        key = lambda s: (s.start, -len(s),)
        return sorted(results, key=key)


def filter_overlapping(results):
    """
    Return a new list from an iterable of `results` discarding contained and
    overlaping Results using these rules:

    - skip a result fully contained in another result.
    - keep the biggest, left-most result of two overlapping results, skip the other.

    For example:
    >>> results = [
    ...     Result(0, 0, 'a'),
    ...     Result(1, 5, 'bcdef'),
    ...     Result(2, 4, 'cde'),
    ...     Result(3, 7, 'defgh'),
    ...     Result(4, 7, 'efgh'),
    ...     Result(8, 9, 'ij'),
    ...     Result(10, 13, 'klmn'),
    ...     Result(11, 15, 'lmnop'),
    ...     Result(16, 16, 'q'),
    ... ]

    >>> expected = [
    ...     Result(0, 0, 'a'),
    ...     Result(1, 5, 'bcdef'),
    ...     Result(8, 9, 'ij'),
    ...     Result(11, 15, 'lmnop'),
    ...     Result(16, 16, 'q'),
    ... ]

    >>> filtered = list(filter_overlapping(results))
    >>> filtered == expected
    True
    """
    results = Result.sort(results)

    # compare pair of results in the sorted sequence: current and next
    i = 0
    while i < len(results) - 1:
        j = i + 1
        while j < len(results):
            curr_res = results[i]
            next_res = results[j]

            logger_debug('curr_res, i, next_res, j:', curr_res, i, next_res, j)
            # disjoint results: break, there is nothing to do
            if next_res.is_after(curr_res):
                logger_debug('  break to next', curr_res)
                break

            # contained result: discard the contained result
            if next_res in curr_res:
                logger_debug('  del next_res contained:', next_res)
                del results[j]
                continue

            # overlap: keep the biggest result and skip the smallest overlapping results
            # in case of length tie: keep the left most
            if curr_res.overlap(next_res):
                if len(curr_res) >= len(next_res):
                    logger_debug('  del next_res smaller overlap:', next_res)
                    del results[j]
                    continue
                else:
                    logger_debug('  del curr_res smaller overlap:', curr_res)
                    del results[i]
                    break
            j += 1
        i += 1
    return results


def add_unmatched(string, results):
    """
    Yield Result object from the original `string` and the search `results` iterable
    of non-overlapping matched substring Result object. New unmatched Results are
    added to the stream for unmatched parts.

    For example:
    >>> string ='abcdefghijklmn'
    >>> results = [
    ...   Result(2, 3, 'cd'),
    ...   Result(7, 7, 'h', None),
    ...   Result(9, 10, 'jk', None),
    ... ]
    >>> expected = [
    ...   Result(0, 1, 'ab'),
    ...   Result(2, 3, 'cd'),
    ...   Result(4, 6, 'efg'),
    ...   Result(7, 7, 'h'),
    ...   Result(8, 8, 'i'),
    ...   Result(9, 10, 'jk'),
    ...   Result(11, 13, 'lmn')
    ... ]
    >>> expected == list(add_unmatched(string, results))
    True

    >>> string ='abc2'
    >>> results = [
    ...   Result(0, 2, 'abc'),
    ... ]
    >>> expected = [
    ...   Result(0, 2, 'abc'),
    ...   Result(3, 3, '2', None),
    ... ]
    >>> expected == list(add_unmatched(string, results))
    True

    """
    string_pos = 0
    for result in Result.sort(results):
        if result.start > string_pos:
            start = string_pos
            end = result.start - 1
            yield Result(start, end, string[start:end + 1])
        yield result
        string_pos = result.end + 1

    len_string = len(string)
    if string_pos < len_string:
        start = string_pos
        end = len_string - 1
        yield Result(start, end, string[start:end + 1])



# license-expression is a free software tool from nexB Inc. and others.
# Visit https://github.com/nexB/license-expression for support and download.
#
# Copyright (c) 2017 nexB Inc. and others. All rights reserved.
# http://nexb.com  and http://aboutcode.org
#
# This software is licensed under the Apache License version 2.0.
#
# You may not use this software except in compliance with the License.
# You may obtain a copy of the License at: http://apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.


from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import sys
from unittest import TestCase

from boolean.boolean import TOKEN_AND
from boolean.boolean import TOKEN_LPAR
from boolean.boolean import TOKEN_OR
from boolean.boolean import TOKEN_RPAR
from boolean.boolean import PARSE_UNBALANCED_CLOSING_PARENS
from boolean.boolean import PARSE_INVALID_SYMBOL_SEQUENCE

from license_expression import TOKEN_WITH
from license_expression import PARSE_INVALID_EXPRESSION
from license_expression import PARSE_INVALID_NESTING

from license_expression import ExpressionError
from license_expression import ExceptionSymbol
from license_expression import Keyword
from license_expression import Licensing
from license_expression import LicenseExpression
from license_expression import LicenseSymbol
from license_expression import LicenseWithExceptionSymbol
from license_expression import ParseError
from license_expression import strip_and_skip_spaces
from license_expression import validate_symbols
from license_expression import group_results_for_with_subexpression


class LicenseSymbolTest(TestCase):

    def test_LicenseSymbol(self):
        sym1 = LicenseSymbol('mit', 'MIT', ['MIT license'], known=True)
        assert sym1 == sym1
        assert 'mit' == sym1.key
        assert 'MIT' == sym1.name
        assert ('MIT license',) == sym1.aliases
        assert sym1.known

        sym2 = LicenseSymbol('mit', 'MIT', ['MIT license'], known=True)
        assert 'mit' == sym2.key
        assert 'MIT' == sym2.name
        assert ('MIT license',) == sym2.aliases
        assert sym2.known
        assert sym1 == sym2
        assert sym1 is not sym2

        sym3 = ExceptionSymbol('mit', 'MIT', ['MIT license'], known=True)
        assert 'mit' == sym3.key
        assert 'MIT' == sym3.name
        assert ('MIT license',) == sym3.aliases
        assert sym3.known
        assert sym1 != sym3

        sym4 = LicenseSymbol('mit', 'MIT', ['MIT license'], known=False)
        assert 'mit' == sym4.key
        assert 'MIT' == sym4.name
        assert ('MIT license',) == sym4.aliases
        assert not sym4.known
        # symbol euqality is based ONLY on the key
        assert sym1 == sym4

        sym5 = LicenseWithExceptionSymbol(sym1, sym3)
        assert sym1 == sym5.license_symbol
        assert sym3 == sym5.exception_symbol

        sym6 = LicenseWithExceptionSymbol(sym4, sym3)
        # symbol euqality is based ONLY on the key
        assert sym5 == sym6


class LicensingTokenizeTest(TestCase):

    def test_tokenize_plain1(self):
        licensing = Licensing()
        expected = [
            (TOKEN_LPAR, '(', 1),
            (LicenseSymbol(key='mit', known=False), u' mit ', 2),
            (TOKEN_RPAR, ')', 7),
            (TOKEN_AND, 'and', 9),
            (LicenseSymbol(key='gpl', known=False), u' gpl', 12)
        ]
        assert expected == list(licensing.tokenize(' ( mit ) and gpl'))

    def test_tokenize_plain2(self):
        licensing = Licensing()
        expected = [
            (TOKEN_LPAR, '(', 0),
            (LicenseSymbol(key='mit', known=False), 'mit ', 1),
            (TOKEN_AND, 'and', 5),
            (LicenseSymbol(key='gpl', known=False), ' gpl', 8),
            (TOKEN_RPAR, ')', 12)
        ]
        assert expected == list(licensing.tokenize('(mit and gpl)'))

    def test_tokenize_plain3(self):
        licensing = Licensing()
        expected = [
            (LicenseSymbol(key='mit', known=False), 'mit ', 0),
            (TOKEN_AND, 'AND', 4),
            (LicenseSymbol(key='gpl', known=False), ' gpl ', 7),
            (TOKEN_OR, 'or', 12),
            (LicenseSymbol(key='gpl', known=False), ' gpl', 14)
        ]
        assert expected == list(licensing.tokenize('mit AND gpl or gpl'))

    def test_tokenize_plain4(self):
        licensing = Licensing()
        expected = [
            (TOKEN_LPAR, '(', 0),
            (TOKEN_LPAR, '(', 1),
            (LicenseSymbol(key='l-a +', known=False), 'l-a + ', 2),
            (TOKEN_AND, 'AND', 8),
            (LicenseSymbol(key='l-b', known=False), ' l-b', 11),
            (TOKEN_RPAR, ')', 15),
            (TOKEN_OR, 'OR', 17),
            (TOKEN_LPAR, '(', 20),
            (LicenseSymbol(key='l -c+', known=False), 'l -c+', 21),
            (TOKEN_RPAR, ')', 26),
            (TOKEN_RPAR, ')', 27)
        ]
        assert expected == list(licensing.tokenize('((l-a + AND l-b) OR (l -c+))'))

    def test_tokenize_plain5(self):
        licensing = Licensing()
        expected = [
            (TOKEN_LPAR, '(', 0),
            (TOKEN_LPAR, '(', 1),
            (LicenseSymbol(key='l-a +', known=False), 'l-a + ', 2),
            (TOKEN_AND, 'AND', 8),
            (LicenseSymbol(key='l-b', known=False), ' l-b', 11),
            (TOKEN_RPAR, ')', 15),
            (TOKEN_OR, 'OR', 17),
            (TOKEN_LPAR, '(', 20),
            (LicenseSymbol(key='l -c+', known=False), 'l -c+', 21),
            (TOKEN_RPAR, ')', 26),
            (TOKEN_RPAR, ')', 27),
            (TOKEN_AND, 'and', 29),
            (LicenseWithExceptionSymbol(
                license_symbol=LicenseSymbol(key='gpl', known=False),
                exception_symbol=ExceptionSymbol(key='classpath exception', known=False)),
             ' gpl  with  classpath exception', 32
            )
        ]
        assert expected == list(licensing.tokenize('((l-a + AND l-b) OR (l -c+)) and gpl with classpath exception'))

    def get_symbols(self):
        gpl_20 = LicenseSymbol('gpl-2.0', 'GPL-2.0', ['The GNU GPL 20'], known=True)
        gpl_20_plus = LicenseSymbol(
            'gpl-2.0+', 'GPL-2.0 or later',
            ['The GNU GPL 20 or later', 'GPL-2.0 or later', 'GPL v2.0 or later'], known=True)
        lgpl_21 = LicenseSymbol('lgpl-2.1', 'LGPL-2.1', ['LGPL v2.1'], known=True)
        mit = LicenseSymbol('mit', 'MIT', ['MIT license'], known=True)
        symbols = [gpl_20, gpl_20_plus, lgpl_21, mit]
        licensing = Licensing(symbols)
        return gpl_20, gpl_20_plus, lgpl_21, mit, licensing

    def test_tokenize_1(self):
        gpl_20, _gpl_20_plus, lgpl_21, mit, licensing = self.get_symbols()
        result = licensing.tokenize('The GNU GPL 20 or LGPL-2.1 and mit')
        expected = [
            (gpl_20, 'The GNU GPL 20', 0),
            (TOKEN_OR, 'or', 15),
            (lgpl_21, 'LGPL-2.1', 18),
            (TOKEN_AND, 'and', 27),
            (mit, 'mit', 31)]
        assert expected == list(result)

    def test_tokenize_with_trailing_unknonw(self):
        gpl_20, _gpl_20_plus, lgpl_21, mit, licensing = self.get_symbols()
        result = licensing.tokenize('The GNU GPL 20 or LGPL-2.1 and mit2')
        expected = [
            (gpl_20, 'The GNU GPL 20', 0),
            (TOKEN_OR, 'or', 15),
            (lgpl_21, 'LGPL-2.1', 18),
            (TOKEN_AND, 'and', 27),
            (mit, 'mit', 31),
            (LicenseSymbol(key=u'2', known=False), u'2', 34)
        ]
        assert expected == list(result)

    def test_tokenize_3(self):
        gpl_20, gpl_20_plus, lgpl_21, mit, licensing = self.get_symbols()
        result = licensing.tokenize('The GNU GPL 20 or later or (LGPL-2.1 and mit) or The GNU GPL 20 or mit 123')
        expected = [
            (gpl_20_plus, 'The GNU GPL 20 or later', 0),
            (TOKEN_OR, 'or', 24),
            (TOKEN_LPAR, '(', 27),
            (lgpl_21, 'LGPL-2.1', 28),
            (TOKEN_AND, 'and', 37),
            (mit, 'mit', 41),
            (TOKEN_RPAR, ')', 44),
            (TOKEN_OR, 'or', 46),
            (gpl_20, 'The GNU GPL 20', 49),
            (TOKEN_OR, 'or', 64),
            (mit, 'mit', 67),
            (LicenseSymbol(key='123', aliases=(), known=False), ' 123', 70)
        ]
        assert expected == list(result)

    def test_tokenize_unknown_as_trailing_single_attached_character(self):
        symbols = [LicenseSymbol('mit', 'MIT', ['MIT license'])]
        l = Licensing(symbols)
        result = list(l.tokenize('mit2'))
        expected = [
            (LicenseSymbol(key='mit', name='MIT', aliases=('MIT license',), known=True), 'mit', 0),
            (LicenseSymbol(key='2', known=False), '2', 3),
        ]
        assert expected == result


class LicensingParseTest(TestCase):

    def test_parse_does_not_raise_error_for_empty_expression(self):
        licensing = Licensing()
        assert None == licensing.parse('')

    def test_parse(self):
        expression = ' ( (( gpl and bsd ) or lgpl)  and gpl-exception) '
        expected = '((gpl AND bsd) OR lgpl) AND gpl-exception'
        licensing = Licensing()
        self.assertEqual(expected, str(licensing.parse(expression)))

    def test_parse_raise_ParseError(self):
        expression = ' ( (( gpl and bsd ) or lgpl)  and gpl-exception)) '
        licensing = Licensing()
        try:
            licensing.parse(expression)
            self.fail('ParseError should be raised')
        except ParseError as pe:
            expected = {'error_code': PARSE_UNBALANCED_CLOSING_PARENS, 'position': 48, 'token_string': ')', 'token_type': TOKEN_RPAR}
            assert expected == _parse_error_as_dict(pe)

    def get_syms_and_licensing(self):
        a = LicenseSymbol('l-a', known=True)
        ap = LicenseSymbol('l-a+', 'L-a+', ['l-a +'], known=True)
        b = LicenseSymbol('l-b', known=True)
        c = LicenseSymbol('l-c', known=True)
        symbols = [a, ap, b, c]
        return a, ap, b, c, Licensing(symbols)

    def test_parse_license_expression1(self):
        a, _ap, _b, _c, licensing = self.get_syms_and_licensing()
        express_string = 'l-a'
        result = licensing.parse(express_string)
        assert express_string == str(result)
        expected = a
        assert expected == result
        assert [] == licensing.unknown_license_keys(result)

    def test_parse_license_expression_with_alias(self):
        _a, ap, _b, _c, licensing = self.get_syms_and_licensing()
        express_string = 'l-a +'
        result = licensing.parse(express_string)
        assert 'l-a+' == str(result)
        expected = ap
        assert expected == result
        assert [] == licensing.unknown_license_keys(result)

    def test_parse_license_expression3(self):
        _a, ap, _b, _c, licensing = self.get_syms_and_licensing()
        express_string = 'l-a+'
        result = licensing.parse(express_string)
        assert express_string == str(result)
        expected = ap
        assert expected == result
        assert [] == licensing.unknown_license_keys(result)

    def test_parse_license_expression4(self):
        _a, _ap, _b, _c, licensing = self.get_syms_and_licensing()
        express_string = '(l-a)'
        result = licensing.parse(express_string)
        assert 'l-a' == str(result)
        expected = LicenseSymbol(key=u'l-a', name=u'l-a', aliases=(), known=True)
        assert expected == result
        assert [] == licensing.unknown_license_keys(result)

    def test_parse_license_expression5(self):
        _a, ap, b, c, licensing = self.get_syms_and_licensing()
        express_string = '((l-a+ AND l-b) OR (l-c))'
        result = licensing.parse(express_string)
        assert '(l-a+ AND l-b) OR l-c' == str(result)
        expected = licensing.OR(licensing.AND(ap, b), c)
        assert expected == result
        assert [] == licensing.unknown_license_keys(result)

    def test_parse_license_expression6(self):
        a, _ap, b, _c, licensing = self.get_syms_and_licensing()
        express_string = 'l-a and l-b'
        result = licensing.parse(express_string)
        assert 'l-a AND l-b' == str(result)
        expected = licensing.AND(a, b)
        assert expected == result
        assert [] == licensing.unknown_license_keys(result)

    def test_parse_license_expression7(self):
        a, _ap, b, _c, licensing = self.get_syms_and_licensing()
        express_string = 'l-a or l-b'
        result = licensing.parse(express_string)
        assert 'l-a OR l-b' == str(result)
        expected = licensing.OR(a, b)
        assert expected == result
        assert [] == licensing.unknown_license_keys(result)

    def test_parse_license_expression8(self):
        a, _ap, b, c, licensing = self.get_syms_and_licensing()
        express_string = 'l-a and l-b OR l-c'
        result = licensing.parse(express_string)
        assert '(l-a AND l-b) OR l-c' == str(result)
        expected = licensing.OR(licensing.AND(a, b), c)
        assert expected == result
        assert [] == licensing.unknown_license_keys(result)

    def test_parse_license_expression8_twice(self):
        _a, _ap, _b, _c, licensing = self.get_syms_and_licensing()
        express_string = 'l-a and l-b OR l-c'
        result = licensing.parse(express_string)
        assert '(l-a AND l-b) OR l-c' == str(result)
        # there was some issues with reusing a Licensing
        result = licensing.parse(express_string)
        assert '(l-a AND l-b) OR l-c' == str(result)

    def test_parse_license_expression_with_trailing_space_plus(self):
        symbols = [
            LicenseSymbol('l-a', known=True),
            LicenseSymbol('l-a+', 'L-a+', ['l-a +'], known=True),
            LicenseSymbol('l-b', known=True),
            LicenseSymbol('l-c', known=True),
        ]
        licensing = Licensing(symbols)

        expresssion_str = 'l-a'
        result = licensing.parse(expresssion_str)
        assert expresssion_str == str(result)
        assert [] == licensing.unknown_license_keys(result)

        # plus sign is not attached to the symbol, but an alias
        expresssion_str = 'l-a +'
        result = licensing.parse(expresssion_str)
        assert 'l-a+' == str(result).lower()
        assert [] == licensing.unknown_license_keys(result)

        expresssion_str = '(l-a)'
        result = licensing.parse(expresssion_str)
        assert 'l-a' == str(result).lower()
        assert [] == licensing.unknown_license_keys(result)

        expresssion_str = '((l-a+ AND l-b) OR (l-c))'
        result = licensing.parse(expresssion_str)
        assert '(l-a+ AND l-b) OR l-c' == str(result)
        assert [] == licensing.unknown_license_keys(result)

        expresssion_str = 'l-a and l-b'
        result = licensing.parse(expresssion_str)
        assert 'l-a AND l-b' == str(result)
        assert [] == licensing.unknown_license_keys(result)

        expresssion_str = 'l-a or l-b'
        result = licensing.parse(expresssion_str)
        assert 'l-a OR l-b' == str(result)
        assert [] == licensing.unknown_license_keys(result)

        expresssion_str = 'l-a and l-b OR l-c'
        result = licensing.parse(expresssion_str)
        assert '(l-a AND l-b) OR l-c' == str(result)
        assert [] == licensing.unknown_license_keys(result)

    def test_parse_invalid_expression_raise_expression(self):
        licensing = Licensing()

        expr = 'wrong'
        licensing.parse(expr)

        expr = 'l-a AND none'
        licensing.parse(expr)

        expr = '(l-a + AND l-b'
        try:
            licensing.parse(expr)
            self.fail("Exception not raised when validating '%s'" % expr)
        except (ExpressionError, ParseError):
            pass

        expr = '(l-a + AND l-b))'
        try:
            licensing.parse(expr)
            self.fail("Exception not raised when validating '%s'" % expr)
        except (ExpressionError, ParseError):
            pass

        expr = 'l-a AND'
        try:
            licensing.parse(expr)
            self.fail("Exception not raised when validating '%s'" % expr)
        except (ExpressionError, ParseError):
            pass

        expr = 'OR l-a'
        try:
            licensing.parse(expr)
            self.fail("Exception not raised when validating '%s'" % expr)
        except (ExpressionError, ParseError):
            pass

        expr = '+ l-a'
        licensing.parse(expr)

    def test_parse_can_parse(self):
        licensing = Licensing()
        expr = licensing.parse(' GPL-2.0 or LGPL 2.1 and mit ')
        gpl2 = LicenseSymbol(key='gpl-2.0', name='GPL-2.0', known=False)
        lgpl = LicenseSymbol(key='lgpl 2.1', name='LGPL 2.1', known=False)
        mit = LicenseSymbol(key='mit', name='mit', known=False)
        expected = [gpl2, lgpl, mit]
        self.assertEqual(expected, licensing.license_symbols(expr))
        self.assertEqual('gpl-2.0 OR (lgpl 2.1 AND mit)', str(expr))

        expected = licensing.OR(gpl2, licensing.AND(lgpl, mit))
        assert expected == expr

    def test_parse_errors_catch_invalid_nesting(self):
        licensing = Licensing()
        try:
            licensing.parse('mit (and LGPL 2.1)')
            self.fail('Exception not raised')
        except ParseError as pe:
            expected = {'error_code': PARSE_INVALID_NESTING, 'position': 4, 'token_string': '(', 'token_type': TOKEN_LPAR}
            assert expected == _parse_error_as_dict(pe)

    def test_parse_errors_catch_invalid_expression_with_bare_and(self):
        licensing = Licensing()
        try:
            licensing.parse('and')
            self.fail('Exception not raised')
        except ParseError as pe:
            expected = {'error_code': PARSE_INVALID_EXPRESSION, 'position':-1, 'token_string': '', 'token_type': None}
            assert expected == _parse_error_as_dict(pe)

    def test_parse_errors_catch_invalid_expression_with_or_and_no_other(self):
        licensing = Licensing()
        try:
            licensing.parse('or that')
            self.fail('Exception not raised')
        except ParseError as pe:
            expected = {'error_code': PARSE_INVALID_EXPRESSION, 'position':-1, 'token_string': '', 'token_type': None}
            assert expected == _parse_error_as_dict(pe)

    def test_parse_errors_catch_invalid_expression_with_empty_parens(self):
        licensing = Licensing()
        try:
            licensing.parse('with ( )this')
            self.fail('Exception not raised')
        except ParseError as pe:
            expected = {'error_code': PARSE_INVALID_EXPRESSION, 'position': 0, 'token_string':  'with', 'token_type': TOKEN_WITH}
            assert expected == _parse_error_as_dict(pe)

    def test_parse_errors_catch_invalid_non_unicode_byte_strings_on_python3(self):
        py2 = sys.version_info[0] == 2
        py3 = sys.version_info[0] == 3

        licensing = Licensing()

        if py2:
            extra_bytes = bytes(chr(0) + chr(12) + chr(255))
            try:
                licensing.parse('mit (and LGPL 2.1)'.encode('utf-8') + extra_bytes)
                self.fail('Exception not raised')
            except ExpressionError as pe:
                assert str(pe).startswith('expression must be one of')
        
        if py3:
            extra_bytes = bytes(chr(0) + chr(12) + chr(255), encoding='utf-8')
            try:
                licensing.parse('mit (and LGPL 2.1)'.encode('utf-8') + extra_bytes)
                self.fail('Exception not raised')
            except ParseError as pe:
                expected =  {'error_code': PARSE_INVALID_NESTING, 'position': 6, 'token_string': '(', 'token_type': TOKEN_LPAR}
                assert expected == _parse_error_as_dict(pe)



    def test_parse_errors_does_not_raise_error_on_plain_non_unicode_raw_string(self):
        # plain non-unicode string does not raise error
        licensing = Licensing()
        x = licensing.parse(r'mit and (LGPL 2.1)')
        self.assertTrue(isinstance(x, LicenseExpression))

    def test_parse_simplify_and_contain_and_equal(self):
        licensing = Licensing()

        expr = licensing.parse(' GPL-2.0 or LGPL 2.1 and mit ')

        expr2 = licensing.parse(' (mit and LGPL 2.1) or GPL-2.0 ')
        self.assertEqual(expr2.simplify(), expr.simplify())
        self.assertEqual(expr2, expr)

        expr3 = licensing.parse('mit and LGPL 2.1')
        self.assertTrue(expr3 in expr2)

    def test_parse_trailing_char_raise_exception(self):
        _gpl2, _gpl2plus, _lgpl, _mit, licensing = self.get_symbols_and_licensing()
        try:
            licensing.parse('The GNU GPL 20 or LGPL-2.1 and mit2')
        except ParseError as pe:
            expected = {'error_code': PARSE_INVALID_SYMBOL_SEQUENCE, 'position': 34,
                        'token_string': '2', 'token_type': LicenseSymbol('2', known=False)}
            assert expected == _parse_error_as_dict(pe)

    def test_parse_expression_with_trailing_unknown_should_raise_exception(self):
        gpl2, gpl2plus, lgpl, mit, licensing = self.get_symbols_and_licensing()
        unknown = LicenseSymbol(key='123', known=False)

        tokens = list(licensing.tokenize('The GNU GPL 20 or later or (LGPL-2.1 and mit) or The GNU GPL 20 or mit 123'))
        expected = [
            (gpl2plus, 'The GNU GPL 20 or later', 0),
            (TOKEN_OR, 'or', 24),
            (TOKEN_LPAR, '(', 27),
            (lgpl, 'LGPL-2.1', 28),
            (TOKEN_AND, 'and', 37),
            (mit, 'mit', 41),
            (TOKEN_RPAR, ')', 44),
            (TOKEN_OR, 'or', 46),
            (gpl2, 'The GNU GPL 20', 49),
            (TOKEN_OR, 'or', 64),
            (mit, 'mit', 67),
            (unknown, ' 123', 70)
        ]
        assert expected == tokens

        try:
            licensing.parse('The GNU GPL 20 or later or (LGPL-2.1 and mit) or The GNU GPL 20 or mit 123')
        except ParseError as pe:
            expected = {'error_code': PARSE_INVALID_SYMBOL_SEQUENCE, 'position': 70,
                        'token_string': ' 123', 'token_type': unknown}
            assert expected == _parse_error_as_dict(pe)

    def test_parse_expression_with_trailing_unknown_should_raise_exception2(self):
        _gpl2, _gpl2plus, _lgpl, _mit, licensing = self.get_symbols_and_licensing()
        unknown = LicenseSymbol(key='123', known=False)
        try:
            licensing.parse('The GNU GPL 20 or mit 123')
        except ParseError as pe:
            expected = {'error_code': PARSE_INVALID_SYMBOL_SEQUENCE, 'position': 21,
                        'token_string': ' 123', 'token_type': unknown}
            assert expected == _parse_error_as_dict(pe)

    def test_parse_expression_with_WITH(self):
        gpl2, gpl2plus, lgpl, mit, licensing = self.get_symbols_and_licensing()
        tokens = list(licensing.tokenize('The GNU GPL 20 or later or (LGPL-2.1 and mit) or The GNU GPL 20 or mit with 123'))
        expected = [
            (gpl2plus, u'The GNU GPL 20 or later', 0),
            (2, u'or', 24),
            (4, u'(', 27),
            (lgpl, u'LGPL-2.1', 28),
            (1, u'and', 37),
            (mit, u'mit', 41),
            (5, u')', 44),
            (2, u'or', 46),
            (gpl2, u'The GNU GPL 20', 49),
            (2, u'or', 64),
            (LicenseWithExceptionSymbol(mit, ExceptionSymbol(key=u'123', known=False)), u'mit with  123', 67)
        ]
        assert expected == tokens

        expr = licensing.parse('The GNU GPL 20 or later or (LGPL-2.1 and mit) or The GNU GPL 20 or mit with 123')
        expected = 'gpl-2.0+ OR (lgpl-2.1 AND mit) OR gpl-2.0 OR mit WITH 123'
        assert expected == str(expr)
        assert ['123'] == licensing.unknown_license_keys(expr)

    def test_parse_of_side_by_side_symbols_raise_exception(self):
        gpl2 = LicenseSymbol('gpl')
        l = Licensing([gpl2])
        try:
            l.parse('gpl mit')
            self.fail('ParseError not raised')
        except ParseError:
            pass

    def test_license_expression_is_equivalent(self):
        is_equivalent = Licensing().is_equivalent

        self.assertTrue(is_equivalent('mit AND gpl', 'mit AND gpl'))
        self.assertTrue(is_equivalent('mit AND gpl', 'gpl AND mit'))
        self.assertTrue(is_equivalent('mit AND gpl and apache', 'apache and gpl AND mit'))
        self.assertTrue(is_equivalent('mit AND (gpl AND apache)', '(mit AND gpl) AND apache'))

        # Real-case example of generated expression vs. stored expression:
        ex1 = '''Commercial
            AND apache-1.1 AND apache-2.0 AND aslr AND bsd-new
            AND cpl-1.0 AND epl-1.0
            AND ibm-icu AND ijg AND jdom AND lgpl-2.1
            AND mit-open-group AND mpl-1.1 AND sax-pd AND unicode AND w3c AND
            w3c-documentation'''

        ex2 = '''
            apache-1.1 AND apache-2.0 AND aslr AND bsd-new
            AND cpl-1.0 AND epl-1.0
            AND lgpl-2.1 AND ibm-icu AND ijg
            AND jdom AND mit-open-group
            AND mpl-1.1 AND Commercial AND sax-pd AND unicode
            AND w3c-documentation AND w3c'''

        self.assertTrue(is_equivalent(ex1, ex2))
        self.assertFalse(is_equivalent('mit AND gpl', 'mit OR gpl'))
        self.assertFalse(is_equivalent('mit AND gpl', 'gpl OR mit'))

    def test_validate_symbols(self):
        symbols = [
            LicenseSymbol('l-a'),
            LicenseSymbol('l-a+', 'L-a+', ['l-a +']),
            ExceptionSymbol('l-a+'),
            LicenseSymbol('l-b'),
            LicenseSymbol('l-c'),
        ]
        warnings, errors = validate_symbols(symbols)
        expectedw = []
        assert expectedw == warnings
        expectede = [
            'Invalid duplicated license key: l-a+.',
            'Invalid duplicated license name: l-a+.'
        ]
        assert expectede == errors

    def test_license_expression_license_keys(self):
        licensing = Licensing()
        self.assertEqual(['mit', 'gpl'], licensing.license_keys(' ( mit ) and gpl'))
        self.assertEqual(['mit', 'gpl'], licensing.license_keys('(mit and gpl)'))
        # these two are surprising for now: this is because the expression is a
        # logical expression so the order may be different on more complex expressions
        self.assertEqual(['mit', 'gpl'], licensing.license_keys('mit AND gpl or gpl'))
        self.assertEqual(['l-a +', 'l-b', 'l -c+'], licensing.license_keys('((l-a + AND l-b) OR (l -c+))'))

    def get_symbols_and_licensing(self):
        gpl2 = LicenseSymbol('gpl-2.0', 'GPL-2.0', ['The GNU GPL 20'])
        gpl2plus = LicenseSymbol('gpl-2.0+', 'GPL-2.0 or later', ['The GNU GPL 20 or later', 'GPL-2.0 or later', 'GPL v2.0 or later'])
        lgpl = LicenseSymbol('lgpl-2.1', 'LGPL-2.1', ['LGPL v2.1'])
        mit = LicenseSymbol('mit', 'MIT', ['MIT license'])
        symbols = [gpl2, gpl2plus, lgpl, mit]
        licensing = Licensing(symbols)
        return gpl2, gpl2plus, lgpl, mit, licensing

    def test_unknown_keys(self):
        _gpl2, _gpl2plus, _lgpl, _mit, licensing = self.get_symbols_and_licensing()
        expr = licensing.parse('The GNU GPL 20 or LGPL-2.1 and mit')
        expected = 'gpl-2.0 OR (lgpl-2.1 AND mit)'
        assert expected == str(expr)
        assert [] == licensing.unknown_license_keys(expr)

    def test_unknown_keys_with_trailing_char(self):
        gpl2, _gpl2plus, lgpl, mit, licensing = self.get_symbols_and_licensing()
        expr = licensing.parse('The GNU GPL 20 or LGPL-2.1 and mitand2')
        expected = [gpl2, lgpl, mit, LicenseSymbol(key='2', known=False)]
        assert expected == licensing.license_symbols(expr)
        assert ['2'] == licensing.unknown_license_keys(expr)

    def test_end_to_end(self):
        # these were formerly doctest portedd to actual real code tests here
        l = Licensing()
        expr = l.parse(' GPL-2.0 or LGPL 2.1 and mit ')
        expected = 'gpl-2.0 OR (lgpl 2.1 AND mit)'
        assert expected == str(expr)

        expected = [
            LicenseSymbol('GPL-2.0', known=False),
            LicenseSymbol('LGPL 2.1', known=False),
            LicenseSymbol('mit', known=False),
        ]
        assert expected == l.license_symbols(expr)

    def test_pretty(self):
        l = Licensing()
        expr = l.parse(' GPL-2.0 or LGPL 2.1 and mit ')

        expected = '''OR(
  LicenseSymbol('gpl-2.0'),
  AND(
    LicenseSymbol('lgpl 2.1'),
    LicenseSymbol('mit')
  )
)'''
        assert expected == expr.pretty()

    def test_simplify_and_contains(self):
        l = Licensing()

        expr = l.parse(' GPL-2.0 or LGPL 2.1 and mit ')
        expr2 = l.parse(' GPL-2.0 or (mit and LGPL 2.1) ')
        assert expr2.simplify() == expr.simplify()
        expr3 = l.parse('mit and LGPL 2.1')
        assert expr3 in expr2

    def test_simplify_and_equivalent_and_contains(self):
        l = Licensing()
        expr2 = l.parse(' GPL-2.0 or (mit and LGPL 2.1) or bsd Or GPL-2.0  or (mit and LGPL 2.1)')
        # note thats simplification does SORT the symbols such that they can
        # eventually be compared sequence-wise. This sorting is based on license key
        expected = 'bsd OR gpl-2.0 OR (lgpl 2.1 AND mit)'
        assert expected == str(expr2.simplify())

        # Two expressions can be compared for equivalence:
        expr1 = l.parse(' GPL-2.0 or (LGPL 2.1 and mit) ')
        assert 'gpl-2.0 OR (lgpl 2.1 AND mit)' == str(expr1)
        expr2 = l.parse(' (mit and LGPL 2.1)  or GPL-2.0 ')
        assert '(mit AND lgpl 2.1) OR gpl-2.0' == str(expr2)
        assert l.is_equivalent(expr1, expr2)

        assert 'gpl-2.0 OR (lgpl 2.1 AND mit)' == str(expr1.simplify())
        assert 'gpl-2.0 OR (lgpl 2.1 AND mit)' == str(expr2.simplify())
        assert expr1.simplify() == expr2.simplify()

        expr3 = l.parse(' GPL-2.0 or mit or LGPL 2.1')
        assert not l.is_equivalent(expr2, expr3)
        expr4 = l.parse('mit and LGPL 2.1')
        assert expr4.simplify() in expr2.simplify()

        assert l.contains(expr2, expr4)

    def test_create_from_python(self):
        # Expressions can be built from Python expressions, using bitwise operators
        # between Licensing objects, but use with caution. The behavior is not as
        # well specified that using text expression and parse

        licensing = Licensing()
        expr1 = (licensing.LicenseSymbol('GPL-2.0')
                 | (licensing.LicenseSymbol('mit')
                    & licensing.LicenseSymbol('LGPL 2.1')))
        expr2 = licensing.parse(' GPL-2.0 or (mit and LGPL 2.1) ')

        assert 'gpl-2.0 OR (lgpl 2.1 AND mit)' == str(expr1.simplify())
        assert 'gpl-2.0 OR (lgpl 2.1 AND mit)' == str(expr2.simplify())

        # WARNING: this is surprising and is due to some type wrapping happening in
        # .py
        assert not licensing.is_equivalent(expr1, expr2)

        a = licensing.OR(
            LicenseSymbol(key=u'gpl-2.0'),
            licensing.AND(LicenseSymbol(key=u'mit'),
                LicenseSymbol(key=u'lgpl 2.1')
                )
            )
        b = licensing.OR(
             LicenseSymbol(key=u'gpl-2.0'),
             licensing.AND(LicenseSymbol(key=u'mit'),
                 LicenseSymbol(key=u'lgpl 2.1')
                 )
            )
        assert a == b

    def test_get_license_symbols(self):
        l = Licensing()
        expected = [
            LicenseSymbol('GPL-2.0', known=False),
            LicenseSymbol('mit', known=False),
            LicenseSymbol('LGPL 2.1', known=False)
        ]
        assert expected == l.license_symbols(' GPL-2.0 and mit or LGPL 2.1 and mit ')

    def test_get_license_symbols2(self):
        l = Licensing()
        expected = [
            LicenseSymbol('GPL-2.0', known=False),
            LicenseSymbol('LATER', known=False),
            LicenseSymbol('mit', known=False),
            LicenseSymbol('LGPL 2.1+', known=False),
            LicenseSymbol('mit', known=False),
            ExceptionSymbol('Foo exception', known=False),
        ]
        expr = ' GPL-2.0 or LATER and mit or LGPL 2.1+ and mit with Foo exception '
        assert expected == l.license_symbols(expr, unique=False)

    def test_get_license_symbols3(self):
        l = Licensing()
        expected = [
            LicenseSymbol('mit', known=False),
            LicenseSymbol('LGPL 2.1+', known=False),
            ExceptionSymbol('Foo exception', known=False),
            LicenseSymbol('GPL-2.0', known=False),
            LicenseSymbol('LATER', known=False),
        ]
        expr = 'mit or LGPL 2.1+ and mit with Foo exception or GPL-2.0 or LATER '
        assert expected == l.license_symbols(expr)

    def test_get_license_symbols4(self):
        l = Licensing()
        expected = [
            LicenseSymbol('GPL-2.0', known=False),
            LicenseSymbol('LATER', known=False),
            ExceptionSymbol('big exception', known=False),
            LicenseSymbol('mit', known=False),
            LicenseSymbol('LGPL 2.1+', known=False),
            LicenseSymbol('later', known=False),
            LicenseSymbol('mit', known=False),
            LicenseSymbol('later', known=False),
            ExceptionSymbol('Foo exception', known=False),
        ]
        expr = (' GPL-2.0 or LATER with big exception and mit or '
                'LGPL 2.1+ or later and mit or later with Foo exception ')
        assert expected == l.license_symbols(expr, unique=False)

    def test_render_complex(self):
        licensing = Licensing()
        expression = '''
        EPL 1.0 AND Apache 1.1 AND Apache 2.0 AND BSD-Modified AND CPL 1.0 AND
        ICU Composite License AND JPEG License AND JDOM License AND LGPL 2.0 AND
        MIT Open Group AND MPL 1.1 AND SAX-PD AND Unicode Inc License Agreement
        AND W3C Software Notice and License AND W3C Documentation License'''

        result = licensing.parse(expression)
        expected = ('EPL 1.0 AND Apache 1.1 AND Apache 2.0 AND BSD-Modified '
        'AND CPL 1.0 AND ICU Composite License AND JPEG License '
        'AND JDOM License AND LGPL 2.0 AND MIT Open Group AND MPL 1.1 '
        'AND SAX-PD AND Unicode Inc License Agreement '
        'AND W3C Software Notice AND License AND W3C Documentation License')

        assert expected == result.render('{name}')
        expectedkey = ('epl 1.0 AND apache 1.1 AND apache 2.0 AND bsd-modified '
        'AND cpl 1.0 AND icu composite license AND jpeg license AND jdom license '
        'AND lgpl 2.0 AND mit open group AND mpl 1.1 AND sax-pd AND '
        'unicode inc license agreement AND w3c software notice AND license '
        'AND w3c documentation license')
        assert expectedkey == result.render('{key}')

    def test_render_with(self):
        licensing = Licensing()
        expression = 'GPL 2.0 with Classpath 2.0 OR BSD-new'
        result = licensing.parse(expression)

        expected = 'GPL 2.0 WITH Classpath 2.0 OR BSD-new'
        assert expected == result.render('{name}')

        expected_html = (
            '<a href="path/gpl 2.0">GPL 2.0</a> '
            'WITH <a href="path/classpath 2.0">Classpath 2.0</a> '
            'OR <a href="path/bsd-new">BSD-new</a>')
        assert expected_html == result.render('<a href="path/{key}">{name}</a>')

        expected = 'gpl 2.0 WITH classpath 2.0 OR bsd-new'
        assert expected == result.render('{key}')

    def test_parse_complex(self):
        licensing = Licensing()
        expression = ' GPL-2.0 or later with classpath Exception and mit or  LPL 2.1 and mit or later '
        result = licensing.parse(expression)
        # this may look weird, but we did not provide symbols hence in "or later",
        # "later" is treated as if it were a license
        expected = 'gpl-2.0 OR (later WITH classpath exception AND mit) OR (lpl 2.1 AND mit) OR later'
        assert expected == result.render('{key}')

        licensing = Licensing()
        expr = licensing.parse(" GPL-2.0 or LGPL 2.1 and mit ")
        expected = [
            LicenseSymbol('GPL-2.0', known=False),
            LicenseSymbol('LGPL 2.1', known=False),
            LicenseSymbol('mit', known=False)
        ]
        assert expected == sorted(licensing.license_symbols(expr))
        expected = 'GPL-2.0 OR (LGPL 2.1 AND mit)'
        assert expected == expr.render('{name}')

    def test_license_symbols(self):
        licensing = Licensing([
            'GPL-2.0 or LATER',
            'classpath Exception',
            'something with else+',
            'mit',
            'LGPL 2.1',
            'mit or later'
        ])

        expr = (' GPL-2.0 or LATER with classpath Exception and mit and '
                'mit with SOMETHING with ELSE+ or LGPL 2.1 and '
                'GPL-2.0 or LATER with classpath Exception and '
                'mit or later or LGPL 2.1 or mit or GPL-2.0 or LATER '
                'with SOMETHING with ELSE+ and lgpl 2.1')

        gpl2plus = LicenseSymbol(key='gpl-2.0 or later')
        cpex = ExceptionSymbol(key='classpath exception')
        someplus = ExceptionSymbol(key='something with else+')
        mitplus = LicenseSymbol(key='mit or later')
        mit = LicenseSymbol(key='mit')
        lgpl = LicenseSymbol(key='lgpl 2.1')
        gpl_with_cp = LicenseWithExceptionSymbol(license_symbol=gpl2plus, exception_symbol=cpex)
        mit_with_some = LicenseWithExceptionSymbol(license_symbol=mit, exception_symbol=someplus)
        gpl2_with_someplus = LicenseWithExceptionSymbol(license_symbol=gpl2plus, exception_symbol=someplus)

        expected = [gpl_with_cp, mit, mit_with_some, lgpl, gpl_with_cp, mitplus, lgpl, mit, gpl2_with_someplus, lgpl]
        assert expected == licensing.license_symbols(expr, unique=False, decompose=False)

        expected = [gpl_with_cp, mit, mit_with_some, lgpl, mitplus, gpl2_with_someplus]
        assert expected == licensing.license_symbols(expr, unique=True, decompose=False)

        expected = [gpl2plus, cpex, mit, mit, someplus, lgpl, gpl2plus, cpex, mitplus, lgpl, mit, gpl2plus, someplus, lgpl]
        assert expected == licensing.license_symbols(expr, unique=False, decompose=True)

        expected = [gpl2plus, cpex, mit, someplus, lgpl, mitplus]
        assert expected == licensing.license_symbols(expr, unique=True, decompose=True)

    def test_primary_license_symbol_and_primary_license_key(self):
        licensing = Licensing([
            'GPL-2.0 or LATER',
            'classpath Exception',
            'mit',
            'LGPL 2.1',
            'mit or later'
        ])

        expr = ' GPL-2.0 or LATER with classpath Exception and mit or LGPL 2.1 and mit or later '
        gpl = LicenseSymbol('GPL-2.0 or LATER')
        cpex = ExceptionSymbol('classpath Exception')
        expected = LicenseWithExceptionSymbol(gpl, cpex)
        assert expected == licensing.primary_license_symbol(expr, decompose=False)
        assert gpl == licensing.primary_license_symbol(expr, decompose=True)
        assert 'gpl-2.0 or later' == licensing.primary_license_key(expr)

        expr = ' GPL-2.0 or later with classpath Exception and mit or LGPL 2.1 and mit or later '
        expected = 'GPL-2.0 or LATER WITH classpath Exception'
        assert expected == licensing.primary_license_symbol(expr, decompose=False).render('{name}')

    def test_tokenize_step_by_step_does_not_munge_trailing_symbols(self):
        licensing = Licensing([
            'GPL-2.0 or LATER',
            'classpath Exception',
            'something with else+',
            'mit',
            'LGPL 2.1',
            'mit or later'
        ])

        expr = (' GPL-2.0 or LATER with classpath Exception and mit and '
                'mit with SOMETHING with ELSE+ or LGPL 2.1 and '
                'GPL-2.0 or LATER with classpath Exception and '
                'mit or later or LGPL 2.1 or mit or GPL-2.0 or LATER '
                'with SOMETHING with ELSE+ and lgpl 2.1')

        gpl2plus = LicenseSymbol(key='gpl-2.0 or later')
        cpex = ExceptionSymbol(key='classpath exception')
        cpex_plain = LicenseSymbol(key='classpath exception')
        someplus = ExceptionSymbol(key='something with else+')
        someplus_plain = LicenseSymbol(key='something with else+')
        mitplus = LicenseSymbol(key='mit or later')
        mit = LicenseSymbol(key='mit')
        lgpl = LicenseSymbol(key='lgpl 2.1')
        gpl2plus_with_cpex = LicenseWithExceptionSymbol(license_symbol=gpl2plus, exception_symbol=cpex)
        mit_with_some = LicenseWithExceptionSymbol(license_symbol=mit, exception_symbol=someplus)
        gpl2plus_with_someplus = LicenseWithExceptionSymbol(license_symbol=gpl2plus, exception_symbol=someplus)

        # fist scan
        scanner = licensing.get_scanner()
        result = list(scanner.scan(expr))
        from license_expression._pyahocorasick import Result, Output
        expected = [
            Result(0, 0, ' ', None),

            Result(1, 16, 'GPL-2.0 or LATER', Output('gpl-2.0 or later', gpl2plus)),
            Result(17, 17, ' ', None),
            Result(18, 21, 'with', Output('with', Keyword(value='with', type=10))),
            Result(22, 22, ' ', None),
            Result(23, 41, 'classpath Exception', Output('classpath exception', cpex_plain)),

            Result(42, 42, ' ', None),
            Result(43, 45, 'and', Output('and', Keyword(value='and', type=1))),
            Result(46, 46, ' ', None),
            Result(47, 49, 'mit', Output('mit', mit)),
            Result(50, 50, ' ', None),
            Result(51, 53, 'and', Output('and', Keyword(value='and', type=1))),
            Result(54, 54, ' ', None),

            Result(55, 57, 'mit', Output('mit', mit)),
            Result(58, 58, ' ', None),
            Result(59, 62, 'with', Output('with', Keyword(value='with', type=10))),
            Result(63, 63, ' ', None),
            Result(64, 83, 'SOMETHING with ELSE+', Output('something with else+', someplus_plain)),

            Result(84, 84, ' ', None),
            Result(85, 86, 'or', Output('or', Keyword(value='or', type=2))),
            Result(87, 87, ' ', None),
            Result(88, 95, 'LGPL 2.1', Output('lgpl 2.1', lgpl)),
            Result(96, 96, ' ', None),
            Result(97, 99, 'and', Output('and', Keyword(value='and', type=1))),
            Result(100, 100, ' ', None),

            Result(101, 116, 'GPL-2.0 or LATER', Output('gpl-2.0 or later', gpl2plus)),
            Result(117, 117, ' ', None),
            Result(118, 121, 'with', Output('with', Keyword(value='with', type=10))),
            Result(122, 122, ' ', None),
            Result(123, 141, 'classpath Exception', Output('classpath exception', cpex_plain)),

            Result(142, 142, ' ', None),
            Result(143, 145, 'and', Output('and', Keyword(value='and', type=1))),
            Result(146, 146, ' ', None),
            Result(147, 158, 'mit or later', Output('mit or later', mitplus)),
            Result(159, 159, ' ', None),
            Result(160, 161, 'or', Output('or', Keyword(value='or', type=2))),
            Result(162, 162, ' ', None),
            Result(163, 170, 'LGPL 2.1', Output('lgpl 2.1', lgpl)),
            Result(171, 171, ' ', None),
            Result(172, 173, 'or', Output('or', Keyword(value='or', type=2))),
            Result(174, 174, ' ', None),
            Result(175, 177, 'mit', Output('mit', mit)),
            Result(178, 178, ' ', None),
            Result(179, 180, 'or', Output('or', Keyword(value='or', type=2))),
            Result(181, 181, ' ', None),

            Result(182, 197, 'GPL-2.0 or LATER', Output('gpl-2.0 or later', gpl2plus)),
            Result(198, 198, ' ', None),
            Result(199, 202, 'with', Output('with', Keyword(value='with', type=10))),
            Result(203, 203, ' ', None),
            Result(204, 223, 'SOMETHING with ELSE+', Output('something with else+', someplus_plain)),

            Result(224, 224, ' ', None),
            Result(225, 227, 'and', Output('and', Keyword(value='and', type=1))),
            Result(228, 228, ' ', None),
            Result(229, 236, 'lgpl 2.1', Output('lgpl 2.1', lgpl))
        ]
        assert expected == result
        assert expected[-1].end + 1 == 237 == sum(len(r.string) for r in result)

        # skip spaces
        result = list(strip_and_skip_spaces(result))
        assert 54 == len(expected)
        assert 27 == len(result)

        # group results
        expected_groups = [
            (Result(1, 16, 'GPL-2.0 or LATER', Output('gpl-2.0 or later', gpl2plus)),
             Result(18, 21, 'with', Output('with', Keyword(value='with', type=10))),
             Result(23, 41, 'classpath Exception', Output('classpath exception', cpex_plain))),

            (Result(43, 45, 'and', Output('and', Keyword(value='and', type=1))),),
            (Result(47, 49, 'mit', Output('mit', LicenseSymbol(key='mit', known=True))),),
            (Result(51, 53, 'and', Output('and', Keyword(value='and', type=1))),),

            (Result(55, 57, 'mit', Output('mit', mit)),
             Result(59, 62, 'with', Output('with', Keyword(value='with', type=10))),
             Result(64, 83, 'SOMETHING with ELSE+', Output('something with else+', someplus_plain))),

            (Result(85, 86, 'or', Output('or', Keyword(value='or', type=2))),),
            (Result(88, 95, 'LGPL 2.1', Output('lgpl 2.1', lgpl)),),
            (Result(97, 99, 'and', Output('and', Keyword(value='and', type=1))),),

            (Result(101, 116, 'GPL-2.0 or LATER', Output('gpl-2.0 or later', gpl2plus)),
             Result(118, 121, 'with', Output('with', Keyword(value='with', type=10))),
             Result(123, 141, 'classpath Exception', Output('classpath exception', cpex_plain))),

            (Result(143, 145, 'and', Output('and', Keyword(value='and', type=1))),),
            (Result(147, 158, 'mit or later', Output('mit or later', mitplus)),),
            (Result(160, 161, 'or', Output('or', Keyword(value='or', type=2))),),
            (Result(163, 170, 'LGPL 2.1', Output('lgpl 2.1', lgpl)),),
            (Result(172, 173, 'or', Output('or', Keyword(value='or', type=2))),),
            (Result(175, 177, 'mit', Output('mit', mit)),),
            (Result(179, 180, 'or', Output('or', Keyword(value='or', type=2))),),

            (Result(182, 197, 'GPL-2.0 or LATER', Output('gpl-2.0 or later', gpl2plus)),
             Result(199, 202, 'with', Output('with', Keyword(value='with', type=10))),
             Result(204, 223, 'SOMETHING with ELSE+', Output('something with else+', someplus_plain))),

            (Result(225, 227, 'and', Output('and', Keyword(value='and', type=1))),),
            (Result(229, 236, 'lgpl 2.1', Output('lgpl 2.1', lgpl)),),

        ]
        result_groups = list(group_results_for_with_subexpression(result))
        assert expected_groups == result_groups

        # finally retest it all with tokenize
        expected = [
            (gpl2plus_with_cpex, 'GPL-2.0 or LATER with classpath Exception', 1),
            (1, 'and', 43),
            (LicenseSymbol(key='mit', known=True), 'mit', 47),
            (1, 'and', 51),
            (mit_with_some, 'mit with SOMETHING with ELSE+', 55),
            (2, 'or', 85),
            (lgpl, 'LGPL 2.1', 88),
            (1, 'and', 97),
            (gpl2plus_with_cpex, 'GPL-2.0 or LATER with classpath Exception', 101),
            (1, 'and', 143),
            (mitplus, 'mit or later', 147),
            (2, 'or', 160),
            (lgpl, 'LGPL 2.1', 163),
            (2, 'or', 172),
            (mit, 'mit', 175),
            (2, 'or', 179),
            (gpl2plus_with_someplus, 'GPL-2.0 or LATER with SOMETHING with ELSE+', 182),
            (1, 'and', 225),
            (lgpl, 'lgpl 2.1', 229),
        ]

        assert expected == list(licensing.tokenize(expr))


def _parse_error_as_dict(pe):
    """
    Return a dict for a ParseError.
    """
    return dict(
        token_type=pe.token_type,
        token_string=pe.token_string,
        position=pe.position,
        error_code=pe.error_code,
    )



#
# license-expression is a free software tool from nexB Inc. and others.
# Visit https://github.com/nexB/license-expression for support and download.
#
# Copyright (c) 2017 nexB Inc. and others. All rights reserved.
# http://nexb.com and http://aboutcode.org
#
# This software is licensed under the Apache License version 2.0.
#
# You may not use this software except in compliance with the License.
# You may obtain a copy of the License at: http://apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.


"""
This module defines a mini language to parse, validate, simplify, normalize and
compare license expressions using a boolean logic engine.

This supports SPDX license expressions and also accepts other license naming
conventions and license identifiers aliases to recognize and normalize licenses.

Using boolean logic, license expressions can be tested for equality, containment,
equivalence and can be normalized or simplified.

The main entry point is the Licensing object.
"""


from __future__ import absolute_import
from __future__ import unicode_literals
from __future__ import print_function

# Python 2 and 3 support
try:
    # Python 2
    unicode
    str = unicode
except NameError:
    # Python 3
    unicode = str

import collections
from copy import copy
from functools import total_ordering
import itertools
import re
import string

import boolean
from boolean import Expression as LicenseExpression

# note these may not all be used here but are imported here to avoid leaking
# boolean.py constants to callers
from boolean.boolean import PARSE_ERRORS
from boolean.boolean import PARSE_INVALID_EXPRESSION
from boolean.boolean import PARSE_INVALID_NESTING
from boolean.boolean import PARSE_INVALID_SYMBOL_SEQUENCE
from boolean.boolean import PARSE_UNBALANCED_CLOSING_PARENS
from boolean.boolean import PARSE_UNKNOWN_TOKEN
from boolean.boolean import ParseError
from boolean.boolean import TOKEN_SYMBOL

from license_expression._pyahocorasick import Trie as Scanner


# append new error codes to PARSE_ERRORS by monkey patching
PARSE_EXPRESSION_NOT_UNICODE = 100
if PARSE_EXPRESSION_NOT_UNICODE not in PARSE_ERRORS:
    PARSE_ERRORS[PARSE_EXPRESSION_NOT_UNICODE] = 'Expression string must be unicode.'

PARSE_INVALID_EXCEPTION_SYMBOL_USAGE = 101
if PARSE_INVALID_EXCEPTION_SYMBOL_USAGE not in PARSE_ERRORS:
    PARSE_ERRORS[PARSE_INVALID_EXCEPTION_SYMBOL_USAGE] = (
        'A license exception symbol can only be used as an exception '
        'in a "WITH exception" statement.')

PARSE_INVALID_SYMBOL_USAGE_AS_EXCEPTION = 102
if PARSE_INVALID_SYMBOL_USAGE_AS_EXCEPTION not in PARSE_ERRORS:
    PARSE_ERRORS[PARSE_INVALID_SYMBOL_USAGE_AS_EXCEPTION] = (
        'A plain license symbol cannot be used as an exception '
        'in a "WITH symbol" statement.')


class ExpressionError(Exception):
    pass


# Used for tokenizing
Keyword = collections.namedtuple('Keyword', 'value type')

# id for "with" token which is not a proper boolean symbol but an expression symbol
TOKEN_WITH = 10

# actual keyword types
_KEYWORDS = [
    Keyword('and', boolean.TOKEN_AND),
    Keyword('or', boolean.TOKEN_OR),
    Keyword('(', boolean.TOKEN_LPAR),
    Keyword(')', boolean.TOKEN_RPAR),
    Keyword('with', TOKEN_WITH),
]

KEYWORDS = tuple(kw.value for kw in _KEYWORDS)


class Licensing(boolean.BooleanAlgebra):
    """
    Define a mini language to parse, validate and compare license expressions.

    For example:

    >>> l = Licensing()
    >>> expr = l.parse(" GPL-2.0 or LGPL 2.1 and mit ")
    >>> expected = 'GPL-2.0 OR (LGPL 2.1 AND mit)'
    >>> assert expected == expr.render('{name}')

    >>> expected = [
    ...   LicenseSymbol('GPL-2.0', known=False),
    ...   LicenseSymbol('LGPL 2.1', known=False),
    ...   LicenseSymbol('mit', known=False)
    ... ]
    >>> assert expected == l.license_symbols(expr)

    >>> symbols = ['GPL-2.0+', 'Classpath', 'BSD']
    >>> l = Licensing(symbols)
    >>> expr = l.parse("GPL-2.0+ with Classpath or (bsd)")
    >>> expected = 'gpl-2.0+ WITH classpath OR bsd'
    >>> assert expected == expr.render('{key}')

    >>> expected = [
    ...   LicenseSymbol('GPL-2.0+', known=True),
    ...   ExceptionSymbol('Classpath', known=True),
    ...   LicenseSymbol('BSD', known=True)
    ... ]
    >>> assert expected == l.license_symbols(expr)
    """
    def __init__(self, symbols=tuple(), strict=False, quiet=True):
        """
        Initialize a Licensing with an optional `symbols` sequence of LicenseSymbol
        or license key strings. If provided and this list data is invalid, raise a
        ValueError.

        If `strict` is True, expressions will be validated against this lits of
        symbols for correctness. In particular if you provided only a list of plain
        strings and not proper ExceptionSymbol for license exception then this will
        tigger an expcetion when an expression is parsed. Otherwise, the list is just
        used as a helper to recognize proper license keys in an expression.
        """
        super(Licensing, self).__init__(Symbol_class=LicenseSymbol, AND_class=AND, OR_class=OR)

        # FIXME: this should be instead a super class of all symbols
        self.LicenseSymbol = self.Symbol


        # True if we will strictly validate the symbol type (e.g. excption or not)
        # using the symbols list.
        self.strict = strict

        if symbols:
            symbols = as_symbols(symbols)
            warns, errors = validate_symbols(symbols)
            if warns and not quiet:
                for w in warns:
                    print(w)
            if errors and not quiet:
                for e in errors:
                    print(e)

            if errors:
                raise ValueError('\n'.join(warns + errors))

        # list of known symbol used for parsing and resolution:
        self.known_symbols = symbols

        # Aho-Corasick automaton-based Scanner used for expression tokenizing
        self.scanner = None

    def build(self, expression, simplify=False):
        """
        Returns an expression from an expression or a string.
        Optionally simplifies the expression if simplify is True.
        Possibly return the expression as-is if this is not a string or an expression.
        """
        expression = self.parse(expression)
        if expression is not None and simplify:
            expression = expression.simplify()
        return expression

    def is_equivalent(self, expression1, expression2):
        """
        Return True if both `expressions` LicenseExpression are equivalent.
        Expressions are either a string or a LicenseExpression object.
        If a string is provided, it will be parsed and simplified.
        """
        ex1 = self.build(expression1, simplify=True)
        ex2 = self.build(expression2, simplify=True)
        if isinstance(ex1, LicenseExpression) and isinstance(ex2, LicenseExpression):
            return ex1 == ex2

    def contains(self, expression1, expression2):
        """
        Return True if expression1 contains expression2.
        Expressions are either a string or a LicenseExpression object.
        If a string is provided, it will be parsed and simplified.
        """
        ex1 = self.build(expression1, simplify=True)
        ex2 = self.build(expression2, simplify=True)
        if isinstance(ex1, LicenseExpression) and isinstance(ex2, LicenseExpression):
            return ex2 in ex1

    def license_symbols(self, expression, unique=True, decompose=True):
        """
        Return a list of LicenseSymbol objects used in an expression in
        the same order as they first appear in the expression tree.

        `expression` is either a string or a LicenseExpression object.
        If a string is provided, it will be parsed.

        If `unique` is True only return unique symbols.

        If `decompose` is True then composite LicenseWithExceptionSymbol instance are
        not returned directly but their underlying license and expection symbols are
        retruned instead.

        For example:
        >>> l = Licensing()
        >>> expected = [
        ...   LicenseSymbol('GPL-2.0', known=False),
        ...   LicenseSymbol('LGPL-2.1+', known=False)
        ... ]
        >>> result = l.license_symbols('GPL-2.0 or LGPL-2.1+')
        >>> assert expected == result
        """
        expression = self.build(expression)
        if not isinstance(expression, LicenseExpression):
            return []

        symbols = (s for s in expression.get_literals() if isinstance(s, BaseSymbol))
        if decompose:
            symbols = itertools.chain.from_iterable(s.decompose() for s in symbols)
        if unique:
            symbols = ordered_unique(symbols)
        return list(symbols)

    def primary_license_symbol(self, expression, decompose=True):
        """
        Return the left-most license symbol of an expression or None.

        If `decompose` is True, only the left-hand license symbol of a decomposed
        LicenseWithExceptionSymbol symbol will be returned if this is the left most
        member. Otherwise a composite LicenseWithExceptionSymbol is returned in this
        case.
        """
        symbols = self.license_symbols(expression, decompose=decompose)
        if symbols:
            return symbols[0]

    def primary_license_key(self, expression):
        """
        Return the left-most license key of an expression or None. The underlying
        symbols are decomposed.
        """
        prim = self.primary_license_symbol(expression, decompose=True)
        if prim:
            return prim.key

    def license_keys(self, expression, unique=True):
        """
        Return a list of licenses keys used in an expression in the same order as
        they first appear in the expression.

        For example:
        >>> l = Licensing()
        >>> expr = ' GPL-2.0 and mit or later with blabla and mit or LGPL 2.1 and mit and mit or later with GPL-2.0'
        >>> expected = ['gpl-2.0', 'mit', 'later', 'blabla', 'lgpl 2.1']
        >>> assert expected == l.license_keys(expr)
        """
        symbols = self.license_symbols(expression, unique=False, decompose=True)
        return self._keys(symbols, unique)

    def _keys(self, symbols, unique=True):
        keys = [ls.key for ls in symbols]
        # note: we only apply this on bare keys strings as we can have the same
        # symbol used as symbol or exception if we are not in strict mode
        if unique:
            keys = ordered_unique(keys)
        return keys

    def unknown_license_symbols(self, expression, unique=True):
        """
        Return a list of unknown licenses symbols used in an expression in the same
        order as they first appear in the expression.
        """
        return [ls for ls in self.license_symbols(expression, unique=unique, decompose=True) if not ls.known]

    def unknown_license_keys(self, expression, unique=True):
        """
        Return a list of unknown licenses keys used in an expression in the same
        order as they first appear in the expression.

        `expression` is either a string or a LicenseExpression object.
        If a string is provided, it will be parsed.

        If `unique` is True only return unique keys.
        """
        symbols = self.unknown_license_symbols(expression, unique=False)
        return self._keys(symbols, unique)

    def parse(self, expression, simplify=False):
        """
        Return a new license LicenseExpression object by parsing a license expression
        string. Check that the expression syntax is valid and raise an Exception,
        ExpressionError or ParseError on errors.
        Return None for empty expressions.

        If `symbols` were provided at Licensing creation time, each license and
        exceptions is recognized from the known licensing symbols. Unknown symbols
        have the `known` flag set to False. Call the`unknown_license_keys` method to get
        unknown license keys.

        If `simplify` is True the returned expression is simplified in a form
        suitable for comparison.

        For example:
        >>> expression = 'EPL 1.0 and Apache 1.1 OR GPL 2.0 with Classpath expection'
        >>> parsed = Licensing().parse(expression)
        >>> expected = '(EPL 1.0 AND Apache 1.1) OR GPL 2.0 WITH Classpath expection'
        >>> assert expected == parsed.render(template='{name}')
        """
        if expression is None:
            return

        expr_types = str, LicenseExpression
        if not isinstance(expression, expr_types):
            if isinstance(expression, bytes):
                try:
                    expression = unicode(expression)
                except:
                    ext = type(expression)
                    raise ExpressionError('expression must be one of %(expr_types)r types and not: %(ext)r' % locals())
            else:
                ext = type(expression)
                raise ExpressionError('expression must be one of %(expr_types)r types and not: %(ext)r' % locals())

        if isinstance(expression, str):
            if not expression or not expression.strip():
                return
            try:
                # this will raise a ParseError on errors
                tokens = list(self.tokenize(expression))
                expression = super(Licensing, self).parse(tokens)
            except TypeError as e:
                raise
                return
                import traceback
                msg = 'Invalid expression syntax: ' + repr(e) + '\n' + traceback.format_exc()
                raise ExpressionError(msg)

        if not isinstance(expression, LicenseExpression):
            raise ExpressionError('expression must be a LicenseExpression once parsed.')

        if simplify:
            expression = expression.simplify()

        return expression

    def tokenize(self, expression):
        """
        Return an iterable of 3-tuple describing each token given an expression
        unicode string. see boolean.BooleanAlgreba.tokenize() for details.

        This 3-tuple contains (token, token string, position):
        - token: either a Symbol instance or one of TOKEN_* token types..
        - token string: the original token unicode string.
        - position: some simple object describing the starting position of the
          original token string in the `expr` string. It can be an int for a
          character offset, or a tuple of starting (row/line, column).
        """
        strict = self.strict
        scanner = self.get_scanner()
        # scan with an automaton, recognize whole symbols+keywords or only keywords
        results = scanner.scan(expression)
        results = strip_and_skip_spaces(results)
        result_groups = group_results_for_with_subexpression(results)
        for group in result_groups:
            if len(group) == 1:
                # a single token
                result = group[0]
                pos = result.start
                token_string = result.string
                output = result.output
                if output:
                    val = output.value
                    if isinstance(val, Keyword):
                        # keyword
                        token = val.type
                        # WITH is not known from the boolean parser as a proper
                        # boolean element so we handle validation ourselves: by
                        # design a single group cannot be a single 'WITH' keyword:
                        # this is an error that we catch and raise here.
                        if token == TOKEN_WITH:
                            raise ParseError(token_type=TOKEN_WITH,
                                             token_string=result.string,
                                             position=result.start,
                                             error_code=PARSE_INVALID_EXPRESSION)

                    elif isinstance(val, LicenseSymbol):
                        if strict and isinstance(val, ExceptionSymbol):
                            raise ParseError(token_type=TOKEN_SYMBOL,
                                             token_string=result.string,
                                             position=result.start,
                                             error_code=PARSE_INVALID_EXCEPTION_SYMBOL_USAGE)

                        # known symbol: create a new copy for sanity and always as a
                        # LicenseSymbol. The strict check above handled possible
                        # errors before.
                        token = LicenseSymbol.from_symbol(val)
                    else:
                        # this should not be possible by design
                        raise Exception('Licensing.tokenize is internally confused...')
                else:
                    # unknown symbol
                    token = LicenseSymbol(result.string, known=False)

            else:
                if len(group) != 3:
                    raise Exception(repr(group))

                assert len(group) == 3
                # this is a A with B seq of three results
                lic_res, _WITH , exc_res = group
                pos = lic_res.start
                token_string = ' '.join([t.string for t in group])

                # known or unknown symbol
                lic_out = lic_res.output
                lic_sym = lic_out and lic_out.value or LicenseSymbol(lic_res.string, known=False)

                if strict and not isinstance(lic_sym, LicenseSymbol):
                    raise ParseError(token_type=TOKEN_SYMBOL,
                                     token_string=lic_res.string,
                                     position=lic_res.start,
                                     error_code=PARSE_INVALID_EXCEPTION_SYMBOL_USAGE)

                lic_sym = LicenseSymbol.from_symbol(lic_sym)

                # known or unknown symbol
                exc_out = exc_res.output
                exc_sym = exc_out and exc_out.value or ExceptionSymbol(exc_res.string, known=False)

                if strict and not isinstance(lic_sym, ExceptionSymbol):
                    raise ParseError(token_type=TOKEN_SYMBOL,
                                     token_string=exc_res.string,
                                     position=exc_res.start,
                                     error_code=PARSE_INVALID_SYMBOL_USAGE_AS_EXCEPTION)

                exc_sym = ExceptionSymbol.from_symbol(exc_sym)

                token = LicenseWithExceptionSymbol(lic_sym, exc_sym)

            yield token, token_string, pos

    def get_scanner(self):
        """
        Return a scanner either cached or created as needed. If symbols were provided
        when this Licensing was created, the scanner will recognize known symbols
        when tokenizing expressions. Otherwise, only keywords are recognized and
        license symbols is anything in between keywords.
        """
        if self.scanner is not None:
            return self.scanner

        self.scanner = scanner = Scanner(ignore_case=True)
        for keyword in _KEYWORDS:
            scanner.add(keyword.value, keyword)

        # only build a basic keyword-only scanner if we have known_symbols
        if not self.known_symbols:
            scanner.make_automaton()
            return scanner

        # self.known_symbols has been created at Licensing initialization time and is
        # already validated and trusted here
        for symbol in self.known_symbols:
            # always use the key even if there are no aliases.
            scanner.add(symbol.key, symbol)
            for alias in symbol.aliases:
                # normalize spaces in aliases. The Scanner will lowercase since we
                # created with ignore_case=True
                if alias:
                    alias = ' '.join(alias.split())
                if alias:
                    scanner.add(alias, symbol)

        scanner.make_automaton()
        return scanner


class Renderable(object):
    """
    An interface for renderable objects.
    """
    def render(self, template='{key}'):
        """
        Return a formatted string rendering for this expression using the `template`
        format string to render each symbol. The variables available are `key` and
        `name` and a custom template can be provided to handle custom HTML rendering
        or similar.

        For symbols that hold multiple licenses (e.g. a WITH statement) the template
        is applied to each symbol individually.
        """
        return NotImplementedError


class BaseSymbol(Renderable, boolean.Symbol):
    """
    A base class for all symbols.
    """

    @classmethod
    def from_symbol(cls, symbol):
        """
        Return a new symbol from that provided symbol.
        """
        raise NotImplementedError

    def decompose(self):
        """
        Yield the underlying symbols of this symbol.
        """
        raise NotImplementedError


#FIXME: we need to implment comparison!!!!
@total_ordering
class LicenseSymbol(BaseSymbol):
    """
    A LicenseSymbol represents a license as used in a license expression.
    """
    def __init__(self, key, name=None, aliases=tuple(), known=True):
        if not key:
            raise ExpressionError(
                'LicenseSymbol key cannot be empty: %(key)r' % locals())

        if not isinstance(key, str):
            if isinstance(key, bytes):
                try:
                    key = unicode(key)
                except:
                    raise ExpressionError(
                        'LicenseSymbol key must be a unicode string: %(key)r' % locals())
            else:
                raise ExpressionError(
                    'LicenseSymbol key must be a unicode string: %(key)r' % locals())

        key = key.strip()

        if not key:
            raise ExpressionError(
                'LicenseSymbol key cannot be blank: %(original_key)r' % locals())

        self.original_key = original_key = key

        # normalize for spaces and make lowercase
        key = ' '.join(key.split())
        self.key = key.lower()

        self.original_name = original_name = name
        if name:
            if not isinstance(name, str):
                if isinstance(name, bytes):
                    try:
                        name = unicode(name)
                    except:
                        raise ExpressionError(
                            'LicenseSymbol name must be a unicode string when provided: '
                            '%(original_name)r' % locals())
                else:
                    raise ExpressionError(
                        'LicenseSymbol name must be a unicode string when provided: '
                        '%(original_name)r' % locals())
            names = name.strip()
            if not names:
                raise ExpressionError(
                    'LicenseSymbol name cannot be blank when provided: '
                    '%(original_name)r' % locals())
            # normalize for spaces
            self.name = ' '.join(names.split())
        else:
            self.name = self.original_key

        self.aliases = aliases and tuple(aliases) or tuple()
        self.known = known

        # super only know about a single "obj" object.
        super(LicenseSymbol, self).__init__(self.key)

    @classmethod
    def from_symbol(cls, symbol):
        assert isinstance(symbol, LicenseSymbol)
        new_symbol = cls(key=symbol.key, name=symbol.name,
                         aliases=symbol.aliases, known=symbol.known)
        new_symbol.original_key = symbol.original_key
        new_symbol.original_name = symbol.original_name
        return new_symbol

    def decompose(self):
        """
        Return an iterable the underlying symbols for this symbol
        """
        yield self

    def render(self, template='{key}'):
        return template.format(**self.__dict__)

    def __hash__(self, *args, **kwargs):
        return hash(self.key)

    def __eq__(self, other):
        return self is other or (isinstance(other, self.__class__) and self.key == other.key)

    __nonzero__ = __bool__ = lambda s: True

    def __str__(self):
        return self.key

    def __repr__(self):
        cls = self.__class__.__name__
        key = self.key
        name = self.original_name and ('name=%r, ' % self.name) or ''
        aliases = self.aliases and ('aliases=%(a)r, ' % {'a': self.aliases}) or ''
        known = self.known
        return '%(cls)s(key=%(key)r, %(name)s%(aliases)sknown=%(known)r)' % locals()


#FIXME: we need to implment comparison!!!!
@total_ordering
class ExceptionSymbol(LicenseSymbol):
    """
    An ExceptionSymbol represents a license exception as used in a license expression.
    """


#FIXME: we need to implment comparison!!!!
@total_ordering
class LicenseWithExceptionSymbol(BaseSymbol):
    """
    A LicenseWithExceptionSymbol represents a license "with" an exception as used in
    a license expression. It holds two LicenseSymbols objects: one for the left-hand
    license proper and one for the right-hand exception to this license and deals
    with the specifics of resolution, validation and representation.
    """
    def __init__(self, license_symbol, exception_symbol):
        if not isinstance(license_symbol, LicenseSymbol):
            raise ExpressionError(
                'license_symbol must be a LicenseSymbol: %(license_symbol)r' % locals())

        if not isinstance(exception_symbol, ExceptionSymbol):
            raise ExpressionError(
                'exception_symbol must be a ExceptionSymbol: %(exception_symbol)r' % locals())

        self.license_symbol = license_symbol
        self.exception_symbol = exception_symbol
        super(LicenseWithExceptionSymbol, self).__init__(str(self))

    @classmethod
    def from_symbol(cls, symbol):
        assert isinstance(symbol, LicenseWithExceptionSymbol)
        return cls(LicenseSymbol.from_symbol(symbol.license_symbol),
                   ExceptionSymbol.from_symbol(symbol.exception_symbol))

    def decompose(self):
        yield self.license_symbol
        yield self.exception_symbol

    def render(self, template='{key}'):
        lic = self.license_symbol.render(template)
        exc = self.exception_symbol.render(template)
        return '%(lic)s WITH %(exc)s' % locals()

    @property
    def known(self):
        return self.license_symbol.known and self.exception_symbol.known

    def __hash__(self, *args, **kwargs):
        return hash((self.license_symbol, self.exception_symbol,))

    def __eq__(self, other):
        return self is other  or (
                isinstance(other, self.__class__)
            and self.license_symbol == other.license_symbol
            and self.exception_symbol == other.exception_symbol)

    __nonzero__ = __bool__ = lambda s: True

    def __str__(self):
        lkey = self.license_symbol.key
        ekey = self.exception_symbol.key
        return '%(lkey)s WITH %(ekey)s' % locals()

    def __repr__(self):
        data = dict(cls=self.__class__.__name__)
        data.update(self.__dict__)
        return '%(cls)s(license_symbol=%(license_symbol)r, exception_symbol=%(exception_symbol)r)' % data


class RenderableFunction(Renderable):
    # derived from the __str__ code in boolean.py
    def render(self, template='{key}'):
        expression_args = self.args
        if len(expression_args) == 1:
            sym = expression_args[0]
            if isinstance(sym, Renderable):
                sym = sym.render(template)
            else:
                # FIXME: CAN THIS REALLY HAPPEN since we only have symbols, or and AND?
                sym = str(sym)
            if self.isliteral:
                rendered = '%s%s' % (self.operator, sym)
            else:
                # NB: the operator str already has a leading and trailing space
                rendered = '%s(%s)' % (self.operator, sym)
            return rendered

        rendered_items = []
        rendered_items_append = rendered_items.append
        for arg in expression_args:
            if isinstance(arg, Renderable):
                rendered = arg.render(template)
            else:
                # FIXME: CAN THIS REALLY HAPPEN since we only have symbols, or and AND?
                rendered = str(arg)

            if arg.isliteral:
                rendered_items_append(rendered)
            else:
                rendered_items_append('(%s)' % rendered)

        return self.operator.join(rendered_items)


class AND(RenderableFunction, boolean.AND):
    """
    Custom representation for the AND operator to uppercase.
    """
    def __init__(self, *args):
        super(AND, self).__init__(*args)
        self.operator = ' AND '


class OR(RenderableFunction, boolean.OR):
    """
    Custom representation for the OR operator to uppercase.
    """
    def __init__(self, *args):
        super(OR, self).__init__(*args)
        self.operator = ' OR '


def ordered_unique(seq):
    """
    Return unique items in a sequence seq preserving the original order.
    """
    if not seq:
        return []
    uniques = []
    for item in seq:
        if item in uniques:
            continue
        uniques.append(item)
    return uniques


def strip_and_skip_spaces(results):
    """
    Yield results given a sequence of Result skipping whitespace-only results
    """
    for result in results:
        if result.string.strip():
            yield result


def group_results_for_with_subexpression(results):
    """
    Yield tuples of (Result) given a sequence of Result such that all symbol-with-
    symbol subsequences of three results are grouped in a tuple and that other
    results are the single res in a tuple.
    """

    # if n-1 is sym, n is with and n+1 is sym: yield this as a group for a with exp
    # otherwise: yield each single result as a group

    results = list(results)

    # check three contiguous result from scanning at a time
    triple_len = 3

    # shortcut if there are no grouping possible
    if len(results) < triple_len:
        for res in results:
            yield (res,)
        return

    # accumulate three contiguous results
    triple = collections.deque()
    triple_popleft = triple.popleft
    triple_clear = triple.clear
    tripple_append = triple.append

    for res in results:
        if len(triple) == triple_len:
            if is_with_subexpression(triple):
                yield tuple(triple)
                triple_clear()
            else:
                prev_res = triple_popleft()
                yield (prev_res,)
        tripple_append(res)

    # end remainders
    if triple:
        if len(triple) == triple_len and is_with_subexpression(triple):
            yield tuple(triple)
        else:
            for res in triple:
                yield (res,)


def is_symbol(result):
    # either the output value is a known sym, por we have no output for unknown sym
    return (result.output and isinstance(result.output.value, LicenseSymbol)) or not result.output


def is_with_keyword(result):
    return (result.output
            and isinstance(result.output.value, Keyword)
            and result.output.value.type == TOKEN_WITH)


def is_with_subexpression(results):
    lic, wit, exc = results
    return (is_symbol(lic) and is_with_keyword(wit) and is_symbol(exc))


def as_symbols(symbols):
    """
    Return a tuple of LicenseSymbol from a sequence of `symbols` or strings. If an
    item is a string, then it will create a LicenseSymbol for it. If this is not a
    string it must be a LicenseSymbol type. It will raise a ValueError expection if
    an item is neither a string or LicenseSymbol.
    """
    if not symbols:
        return symbols
    new_symbols = []
    new_symbols_append = new_symbols.append
    for symbol in symbols:
        if not symbol:
            continue
        if isinstance(symbol, bytes):
            try:
                symbol = unicode(symbol)
            except:
                raise ValueError('%(symbol)r is not a unicode string or a LicenseSymbol instance.' % locals())

        if isinstance(symbol, unicode):
            if symbol.strip():
                new_symbols_append(LicenseSymbol(symbol))
        elif isinstance(symbol, LicenseSymbol):
            new_symbols_append(symbol)
        else:
            raise ValueError('%(symbol)r is not a unicode string or a LicenseSymbol instance.' % locals())
    return tuple(new_symbols)


_valid_key = re.compile('^[A-Za-z0-9\+\-\_\.\:]*$', re.IGNORECASE).match


def is_valid_key(k, _keywords=KEYWORDS):
    return _valid_key(k) and k.lower() not in _keywords


def validate_symbols(symbols, validate_keys=False, _keywords=KEYWORDS):
    """
    Return a tuple of (`warnings`, `errors`) given a list of `symbols` LicenseSymbol
    objects. LicenseSymbol are updated in place ass needed with side-effects.

    - warnings is a list of validation warnings messages (possibly empty if there
      were no warnings).
    - errors is a list of validation error messages (possibly empty if there were no
      errors).

    Keys, names and aliases are cleaned and validated for uniqueness.
    """

    # collection used for checking unicity and correctness
    seen_keys = set()
    seen_names = set()
    seen_aliases = {}
    seen_exceptions = set()

    # collections to accumulate invalid data and build error messages at the end
    not_symbol_classes = []
    invalid_keys = set()
    dupe_keys = set()
    dupe_names = set()
    dupe_exceptions = set()
    dupe_aliases = collections.defaultdict(list)
    invalid_names_as_kw = set()
    invalid_keys_as_kw = set()
    invalid_alias_as_kw = set()
    symbol_not_known = set()

    # warning
    warning_dupe_aliases = set()

    for symbol in symbols:
        if not isinstance(symbol, LicenseSymbol):
            not_symbol_classes.append(symbol)
            continue

        key = symbol.key
        if validate_keys and not is_valid_key(key):
            invalid_keys.add(key)

        key = key.strip()
        keyl = key.lower()

        # ensure keys are unique
        if keyl in seen_keys:
            dupe_keys.add(key)

        # key cannot be an expression keyword
        if keyl in _keywords:
            invalid_keys_as_kw.add(key)

        name = symbol.name
        if not name or not name.strip():
            # use the key (not lowercased) as a default if there is no name
            name = key

        # normalize spaces
        name = ' '.join(name.split())
        namel = name.lower()
        # ensure names are unique
        if namel in seen_names:
            dupe_names.add(name)
        seen_names.add(namel)

        # name cannot be an expression keyword
        if namel in _keywords:
            invalid_names_as_kw.add(name)

        # keep a set of unique seen keys
        seen_keys.add(keyl)

        aliases = symbol.aliases or []
        initial_alias_len = len(aliases)

        # always normalize aliases for spaces and case. keep uniques only
        aliases = set([' '.join(alias.lower().strip().split()) for alias in symbol.aliases])
        # remove empties
        aliases = set(a for a in aliases if a)

        # issue a warning when there are duplicated or empty aliases
        if len(aliases) != initial_alias_len:
            warning_dupe_aliases.add(key)

        # always use a lowercase key as an alias
        aliases.add(keyl)

        # make hashable
        aliases = tuple(sorted(aliases))

        for alias in aliases:
            # note that we do not treat as an error the presence of a duplicated
            # alias pointing to the same key

            # ensure that a possibly duplicated alias does not point to another key
            aliased_key = seen_aliases.get(alias)
            if aliased_key and aliased_key != keyl:
                dupe_aliases[alias].append(key)

            # an alias cannot be an expression keyword
            if alias in _keywords:
                invalid_alias_as_kw.add(alias)

            seen_aliases[alias] = keyl

        if isinstance(symbol, ExceptionSymbol):
            if keyl in seen_exceptions:
                dupe_exceptions.add(keyl)
            else:
                seen_exceptions.add(keyl)

        # make sure to mark the symbol as known
        if not symbol.known:
            symbol_not_known.add(keyl)

    # build warning and error messages from invalid data
    errors = []
    for ind in sorted(not_symbol_classes):
        errors.append('Invalid license reference object missing a key, name, '
                      'aliases or exception attribute: %(ind)s.' % locals())

    for invak in sorted(invalid_keys):
        errors.append('Invalid license key. Can only contain ASCII letters '
                      'and digits and ".+_-": %(invak)s.' % locals())

    for dupe in sorted(dupe_keys):
        errors.append('Invalid duplicated license key: %(dupe)s.' % locals())

    for nm in sorted(dupe_names):
        errors.append('Invalid duplicated license name: %(nm)s.' % locals())

    for dalias, dkeys in sorted(dupe_aliases.items()):
        dkeys = ', '.join(dkeys)
        errors.append('Invalid duplicated alias pointing to different seen_keys: '
                      '%(dalias)s to keys: %(dkeys)s.' % locals())

    for dupe in sorted(dupe_exceptions):
        errors.append('Invalid duplicated license exception key: %(dupe)s.' % locals())

    for ikw in sorted(invalid_keys_as_kw):
        errors.append('Invalid key: cannot be an expression keyword: %(ikw)s.' % locals())

    for ikw in sorted(invalid_names_as_kw):
        errors.append('Invalid name: cannot be an expression keyword: %(ikw)r.' % locals())

    for ikw in sorted(invalid_alias_as_kw):
        errors.append('Invalid alias: cannot be an expression keyword: %(ikw)r.' % locals())

    for us in sorted(symbol_not_known):
        errors.append('Invalid symbol: is not marked as a known symbol: %(us)r.' % locals())

    warnings = []
    for dupeal in sorted(dupe_aliases):
        errors.append('Duplicated or empty aliases ignored for license key: %(dupeal)r.' % locals())

    return warnings, errors



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from __future__ import absolute_import
from __future__ import print_function

from glob import glob
from os.path import basename
from os.path import splitext
from setuptools import find_packages
from setuptools import setup


desc = ('license-expression is small utility library to parse, compare, '
        'simplify and normalize license expressions (such as SPDX license '
        'expressions) using boolean logic.')

setup(
    name='license-expression',
    version='0.12',
    license='apache-2.0',
    description=desc,
    long_description=desc,
    author='nexB Inc.',
    author_email='info@nexb.com',
    url='https://github.com/nexB/license-expression',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Utilities',
    ],
    keywords=[
        'license', 'spdx', 'license expression', 'open source', 'boolean',
        'parse expression', 'normalize expression', 'compare expression',
        'licence'
    ],
    install_requires=[
        'boolean.py >= 3.3, < 4.0.0',
    ]
)
